from ibis import _

from phenex.phenotypes.phenotype import Phenotype
from phenex.filters.relative_time_range_filter import RelativeTimeRangeFilter
from phenex.filters import DateFilter, ValueFilter
from phenex.tables import is_phenex_code_table, PHENOTYPE_TABLE_COLUMNS, PhenotypeTable
from phenex.aggregators import First, Last

from phenex.util import create_logger

logger = create_logger(__name__)


class EventCountPhenotype(Phenotype):
    """
    EventCountPhenotype counts the number of events that occur on distinct days. It is additionally able to filter patients based on:
    1. the number of distinct days an event occurred, by setting value_filter
    2. the number of days between pairs of events

    EventCountPhenotype is a composite phenotype, meaning that it does not directly operate on source data and takes a phenotype as an argument. The phenotype passed to EventCountPhenotype must have return_date set to 'all' (if return_date on the provided phenotype is set to `first` or `last`, there will only be one event per patient...)


    DATE: The event date selected based on `component_date_select` and `return_date` parameters. `return_date` returns multiple rows per patient for all events that fulfill criteria. `return_date` first is the first fulfilling event date, last the last. If component_date_select = 'first' the returned date is a pair of events, if component_date_select = 'second' we return the second of a pair of events.
    VALUE: The number of days that the phenotype of interest has occurred i.e. if 4, that means the phenotype has occurred on 4 distinct days.

    Parameters:
        name: The name of the phenotype.
        phenotype: The phenotype that returns events of interest. Note that return_date must be set to `all` or an error will be thrown.
        value_filter: Set the minimum and/or maximum number of distinct days on which an event may occur.
        relative_time_range: Set the minimum and/or maximum number of days that are allowed to occur between any pair of events.
        return_date: Specifies whether to return the 'first', 'last', or 'all' dates on which the criteria are fulfilled. Default is 'first'.
        component_date_select: Specifies whether to return the 'first' or 'second' event date within each pair of events. Default is 'second'. It is highly recommended to never use 'first', as there is a high risk of introducing immortal time bias.

    Example:
        ```python
        codelist = Codelist(name="example_codelist", codes=[...])

        phenotype = CodelistPhenotype(
            name="example_phenotype",
            domain="CONDITION_OCCURRENCE",
            codelist=codelist,
            return_date='first'
        )

        tables = {"CONDITION_OCCURRENCE": example_code_table}
        multiple_occurrences = EventCountPhenotype(
            phenotype=phenotype,
            value_filter=ValueFilter(min_value=GreaterThanOrEqualTo(2)),
            relative_time_range=RelativeTimeRangeFilter(
                min_days=GreaterThanOrEqualTo(90),
                max_days=LessThanOrEqualTo(180)
            ),
            return_date='first',
            component_date_select='second'
        )

        result_table = multiple_occurrences.execute(tables)
        display(result_table)
        ```
    """

    def __init__(
        self,
        phenotype: Phenotype,
        value_filter: ValueFilter = None,
        relative_time_range: RelativeTimeRangeFilter = None,
        return_date="first",
        component_date_select="second",
        **kwargs,
    ):
        super(EventCountPhenotype, self).__init__(**kwargs)
        self.relative_time_range = relative_time_range
        self.return_date = return_date
        self.component_date_select = component_date_select
        if self.component_date_select not in ["first", "second"]:
            raise ValueError(
                f"Invalid component_date_select: {self.component_date_select}"
            )
        self.value_filter = value_filter
        self.phenotype = phenotype
        self.children = [phenotype]

    def _execute(self, tables) -> PhenotypeTable:
        # Execute the child phenotype to get the initial table to filter
        if self.phenotype.return_date != "all":
            raise ValueError(
                "EventCountPhenotype requires that return_date is set to all on its component phenotype"
            )
        table = self.phenotype.table

        # Select only distinct dates:
        table = table.select(["PERSON_ID", "EVENT_DATE"]).distinct()

        # Count occurrences per PERSON_ID
        occurrence_counts_table = table.group_by("PERSON_ID").aggregate(VALUE=_.count())
        table, occurrence_counts_table = self._perform_value_filtering(
            table, occurrence_counts_table
        )
        table = self._perform_relative_time_range_filtering(table)
        table = self._perform_date_selection(table)
        table = table.left_join(
            occurrence_counts_table.select("PERSON_ID", "VALUE"),
            table.PERSON_ID == occurrence_counts_table.PERSON_ID,
        ).select("PERSON_ID", "EVENT_DATE", "VALUE")

        table = table.mutate(BOOLEAN=True).distinct()
        return table

    def _perform_value_filtering(self, table, occurrence_counts_table):
        if self.value_filter is not None:
            occurrence_counts_table = self.value_filter.filter(occurrence_counts_table)
            table = table.right_join(
                occurrence_counts_table,
                table.PERSON_ID == occurrence_counts_table.PERSON_ID,
            ).select(["PERSON_ID", "EVENT_DATE", "VALUE"])
        return table, occurrence_counts_table

    def _perform_relative_time_range_filtering(self, table):
        if self.relative_time_range is not None:
            # make sure that the 'when' keyword parameter is correctly set to after
            self.relative_time_range.when = "after"
            # Self join and rename event_date columns;
            # the first dates will be called INDEX_DATE
            # the second dates will be called EVENT_DATE
            first_table = table.select(
                "PERSON_ID",
                table.EVENT_DATE.name("INDEX_DATE"),
            )
            second_table = table.select(
                "PERSON_ID",
                table.EVENT_DATE.name("EVENT_DATE"),
            )
            table = first_table.join(
                second_table, first_table.PERSON_ID == second_table.PERSON_ID
            )

            table = table.filter(table.INDEX_DATE <= table.EVENT_DATE)
            # perform relative time range filtering; the first date is the anchor ('index_date')
            table = self.relative_time_range.filter(table)

            if self.component_date_select == "first":
                table = table.select("PERSON_ID", "INDEX_DATE").rename(
                    {"EVENT_DATE": "INDEX_DATE"}
                )
            elif self.component_date_select == "second":
                table = table.select("PERSON_ID", "EVENT_DATE")
        return table

    def _perform_date_selection(self, table, reduce=True):
        if self.return_date is None or self.return_date == "all":
            return table
        if self.return_date == "first":
            aggregator = First(reduce=reduce)
        elif self.return_date == "last":
            aggregator = Last(reduce=reduce)
        else:
            raise ValueError(f"Unknown return_date: {self.return_date}")
        table = aggregator.aggregate(table)
        return table.select("PERSON_ID", "EVENT_DATE")



from typing import Dict
from ibis.expr.types.relations import Table


class DerivedTable:
    """
    A DerivedTable takes one or more source tables and creates a single output table. Possible use cases are:
    1. combining multiple tables into one e.g. taking the vaccination table, patient reported medication, administered medication and prescription table and creating a single medication table.
    2. performing cleaning operations on a table, e.g. removing duplicates, filling in missing values, etc.
    3. performing aggregations on a table, e.g. calculating the number of visits per patient, etc.

    To subclass:
        1. implement execute method, and return a single table
    """

    def __init__(self, source_domain: str, dest_domain: str):
        self.source_domain = source_domain
        self.dest_domain = dest_domain

    def execute(
        self,
        tables: Dict[str, Table],
    ) -> "PhenexTable":
        raise NotImplementedError



from .categorical_filter import CategoricalFilter
from .codelist_filter import CodelistFilter
from .relative_time_range_filter import RelativeTimeRangeFilter
from .value_filter import ValueFilter
from .date_filter import DateFilter, Before, BeforeOrOn, After, AfterOrOn, Date
from .value import (
    GreaterThan,
    GreaterThanOrEqualTo,
    LessThan,
    LessThanOrEqualTo,
    EqualTo,
    Value,
)



from datetime import datetime
from phenex.codelists import *
from phenex.phenotypes import *
from phenex.phenotypes.phenotype import ComputationGraph
from phenex.filters import *
import inspect
from phenex.util import create_logger
from phenex.util.serialization.to_dict import get_phenex_init_params

logger = create_logger(__name__)


def from_dict(data: dict):
    """
    Method to decode all PhenEx classes. Given encoded PhenEx data, it will return the corresponding PhenEx class.
    """
    # logger.debug(f"Decoding data: {data}")

    class_name = data.pop("class_name")
    # logger.debug(f"Class name: {class_name}")
    cls = globals()[class_name]
    all_params = get_phenex_init_params(cls)
    # logger.debug(f"Current params: {all_params}")

    init_args = {}
    kwargs = {}
    for param in all_params:
        if param == "kwargs":
            continue
        if param != "self":
            value = data.get(param)
            param_type = all_params[param].annotation
            # logger.debug(f"Processing param: {param}, value: {value}, type: {param_type}")
            if value is None:
                init_args[param] = None
            elif isinstance(value, list):
                init_args[param] = [
                    (
                        from_dict(item)
                        if isinstance(item, dict) and "class_name" in item.keys()
                        else item
                    )
                    for item in value
                ]
            elif isinstance(value, dict) and "__datetime__" in value:
                init_args[param] = datetime.fromisoformat(value["__datetime__"])
            elif isinstance(value, dict) and "class_name" in value.keys():
                init_args[param] = from_dict(value)
            elif isinstance(value, dict):
                init_args[param] = convert_null_keys_to_none_in_dictionary(value)
            else:
                init_args[param] = value

    # logger.debug(f"Init args: {init_args}")
    # logger.debug(f"Kwargs: {kwargs}")
    if len(kwargs.keys()) > 0:
        return cls(**init_args)
    return cls(**init_args)


def convert_null_keys_to_none_in_dictionary(_dict):
    """
    Given a dictionary with strings 'null' as keys, replaces the 'null' string key with a python NoneType this is required because Codelists are implemented as a dictionary with keys = code_type and If code_type is not defined the key is None (in python) and null in json
    """
    new_dict = {}
    for k, v in _dict.items():
        if k == "null":
            new_key = None
        else:
            new_key = k
        new_dict[new_key] = v
    return new_dict



from typing import Union, List, Optional
from phenex.phenotypes.phenotype import Phenotype
from phenex.filters.codelist_filter import CodelistFilter
from phenex.filters.relative_time_range_filter import RelativeTimeRangeFilter
from phenex.filters.date_filter import DateFilter
from phenex.aggregators import First, Last
from phenex.codelists import Codelist
from phenex.tables import is_phenex_code_table, PHENOTYPE_TABLE_COLUMNS, PhenotypeTable
from phenex.phenotypes.functions import select_phenotype_columns
from ibis import _


class CodelistPhenotype(Phenotype):
    """
    CodelistPhenotype extracts patients from a CodeTable based on a specified codelist and other optional filters such as date range, relative time range and categorical filters.

    Parameters:
        domain: The domain of the phenotype.
        codelist: The codelist used for filtering.
        name: The name of the phenotype. Optional. If not passed, name will be derived from the name of the codelist.
        date_range: A date range filter to apply.
        relative_time_range: A relative time range filter or a list of filters to apply.
        return_date: Specifies whether to return the 'first', 'last', or 'nearest' event date. Default is 'first'.
        categorical_filter: Additional categorical filters to apply.

    Attributes:
        table (PhenotypeTable): The resulting phenotype table after filtering (None until execute is called)

    Examples:

    Example: Inpatient Atrial Fibrillation (OMOP)
        ```python
        from phenex.phenotypes import CodelistPhenotype
        from phenex.codelists import Codelist
        from phenex.mappers import OMOPDomains
        from phenex.filters import DateFilter, CategoricalFilter, Value
        from phenex.ibis_connect import SnowflakeConnector

        con = SnowflakeConnector() # requires some configuration
        mapped_tables = OMOPDomains.get_mapped_tables(con)

        af_codelist = Codelist([313217]) # list of concept ids
        date_range = DateFilter(
            min_date=After("2020-01-01"),
            max_date=Before("2020-12-31")
            )

        inpatient = CategoricalFilter(
            column_name='VISIT_DETAIL_CONCEPT_ID',
            allowed_values=[9201],
            domain='VISIT_DETAIL'
        )

        af_phenotype = CodelistPhenotype(
            name="af",
            domain='CONDITION_OCCURRENCE',
            codelist=af_codelist,
            date_range=date_range,
            return_date='first',
            categorical_filter=inpatient
        )

        af = af_phenotype.execute(mapped_tables)
        af.head()
        ```

    Example: Myocardial Infarction One Year Pre-index (OMOP)
        ```python
        from phenex.filters import RelativeTimeRangeFilter, Value

        af_phenotype = (...) # take from above example

        oneyear_preindex = RelativeTimeRangeFilter(
            min_days=Value('>', 0), # exclude index date
            max_days=Value('<', 365),
            anchor_phenotype=af_phenotype # use af phenotype above as reference date
            )

        mi_codelist = Codelist([49601007]) # list of concept ids
        mi_phenotype = CodelistPhenotype(
            name='mi',
            domain='CONDITION_OCCURRENCE',
            codelist=mi_codelist,
            return_date='first',
            relative_time_range=oneyear_preindex
        )
        mi = mi_phenotype.execute(mapped_tables)
        mi.head()
        ```
    """

    def __init__(
        self,
        domain: str,
        codelist: Codelist,
        name: Optional[str] = None,
        date_range: DateFilter = None,
        relative_time_range: Union[
            RelativeTimeRangeFilter, List[RelativeTimeRangeFilter]
        ] = None,
        return_date="first",
        categorical_filter: Optional["CategoricalFilter"] = None,
        **kwargs,
    ):
        if name is None:
            name = codelist.name
        super(CodelistPhenotype, self).__init__(name=name, **kwargs)

        self.codelist_filter = CodelistFilter(codelist)
        self.codelist = codelist
        self.categorical_filter = categorical_filter
        self.date_range = date_range
        self.return_date = return_date
        assert self.return_date in [
            "first",
            "last",
            "nearest",
            "all",
        ], f"Unknown return_date: {return_date}"
        self.table: PhenotypeTable = None
        self.domain = domain
        if isinstance(relative_time_range, RelativeTimeRangeFilter):
            relative_time_range = [relative_time_range]

        self.relative_time_range = relative_time_range
        if self.relative_time_range is not None:
            for rtr in self.relative_time_range:
                if rtr.anchor_phenotype is not None:
                    self.children.append(rtr.anchor_phenotype)

    def _execute(self, tables) -> PhenotypeTable:
        code_table = tables[self.domain]
        code_table = self._perform_codelist_filtering(code_table)
        code_table = self._perform_categorical_filtering(code_table, tables)
        code_table = self._perform_time_filtering(code_table)
        code_table = self._perform_date_selection(code_table)
        return select_phenotype_columns(code_table)

    def _perform_codelist_filtering(self, code_table):
        assert is_phenex_code_table(code_table)
        code_table = self.codelist_filter.filter(code_table)
        return code_table

    def _perform_categorical_filtering(self, code_table, tables):
        if self.categorical_filter is not None:
            assert is_phenex_code_table(code_table)
            code_table = self.categorical_filter.autojoin_filter(code_table, tables)
        return code_table

    def _perform_time_filtering(self, code_table):
        if self.date_range is not None:
            code_table = self.date_range.filter(code_table)
        if self.relative_time_range is not None:
            for rtr in self.relative_time_range:
                code_table = rtr.filter(code_table)
        return code_table

    def _perform_date_selection(self, code_table, reduce=True):
        if self.return_date is None or self.return_date == "all":
            return code_table
        if self.return_date == "first":
            aggregator = First(reduce=reduce)
        elif self.return_date == "last":
            aggregator = Last(reduce=reduce)
        else:
            raise ValueError(f"Unknown return_date: {self.return_date}")
        return aggregator.aggregate(code_table)

    def get_codelists(self) -> List[Codelist]:
        """
        Get all codelists used in the phenotype definition, including all children / dependent phenotypes.

        Returns:
            codeslist: A list of codelists used in the cohort definition.
        """
        codelists = [self.codelist]
        for p in self.children:
            codelists.extend(p.get_codelists())
        return codelists



import ibis

from phenex.codelists import Codelist
from phenex.tables import CodeTable, is_phenex_code_table
from phenex.filters.filter import Filter
from typing import List, Tuple
import pandas as pd


class CodelistFilter(Filter):
    """
    CodelistFilter is a class designed to filter a CodeTable based on a specified codelist.

    Attributes:
        codelist (Codelist): The codelist used for filtering the CodeTable.
        name (str): The name of the filter. Defaults to the name of the codelist if not provided.
        use_code_type (bool): A flag indicating whether to use the code type in the filtering process. Defaults to True.
    """

    def __init__(self, codelist: Codelist, name=None):
        self.codelist = codelist
        self.name = name or self.codelist.name
        self.codelist_as_tuples = self._convert_codelist_to_tuples()
        super(CodelistFilter, self).__init__()

    def _convert_codelist_to_tuples(self) -> List[Tuple[str, str]]:
        if self.codelist is not None:
            if not isinstance(self.codelist, Codelist):
                raise ValueError("Codelist must be an instance of Codelist")
            return [
                (ct, c)
                for ct, codes in self.codelist.resolved_codelist.items()
                for c in codes
            ]
        return []

    def _filter(self, code_table: CodeTable) -> CodeTable:
        assert is_phenex_code_table(code_table)

        if self.codelist.fuzzy_match:
            return self._filter_fuzzy_codelist(code_table)
        else:
            return self._filter_literal_codelist(code_table)

    def _filter_fuzzy_codelist(self, code_table):
        filter_condition = False
        for code_type, codelist in self.codelist.resolved_codelist.items():
            codelist = [str(code) for code in codelist]
            if self.codelist.use_code_type:
                filter_condition = filter_condition | (
                    (code_table.CODE_TYPE == code_type)
                    & (code_table.CODE.cast("str").like(codelist))
                )
            else:
                filter_condition = filter_condition | code_table.CODE.cast("str").like(
                    codelist
                )

        filtered_table = code_table.filter(filter_condition)
        return filtered_table

    def _filter_literal_codelist(self, code_table):
        # Generate the codelist table as an Ibis literal set
        codelist_df = pd.DataFrame(
            self.codelist_as_tuples, columns=["code_type", "code"]
        ).fillna("")
        codelist_table = ibis.memtable(codelist_df)

        # Create a join condition based on code and possibly code_type
        code_column = code_table.CODE
        if self.codelist.use_code_type:
            code_type_column = code_table.CODE_TYPE
            join_condition = (
                code_column.cast("str") == codelist_table.code.cast("str")
            ) & (code_type_column == codelist_table.code_type)
        else:
            join_condition = code_column.cast("str") == codelist_table.code.cast("str")

        # return table with downselected columns, of same type as input table
        filtered_table = code_table.inner_join(codelist_table, join_condition).select(
            code_table.columns
        )
        return filtered_table



from typing import Dict, Optional, Union, List
from phenex.filters.relative_time_range_filter import RelativeTimeRangeFilter
from ibis.expr.types.relations import Table
from phenex.phenotypes.phenotype import Phenotype
from phenex.tables import PhenotypeTable, is_phenex_person_table
import ibis


class DeathPhenotype(Phenotype):
    """
    DeathPhenotype is a class that represents a death-based phenotype. It filters individuals
    who have died and returns their date of death.

    Parameters:
        name: Name of the phenotype, default is 'death'.
        domain: Domain of the phenotype, default is 'PERSON'.
        relative_time_range: Filter patients relative to some date (e.g. death after discharge from hospital)

    Attributes:
        table (PhenotypeTable): The resulting phenotype table after filtering (None until execute is called)
    """

    def __init__(
        self,
        name: Optional[str] = "DEATH",
        domain: str = "PERSON",
        relative_time_range: Union[
            RelativeTimeRangeFilter, List[RelativeTimeRangeFilter]
        ] = None,
        **kwargs
    ):
        self.domain = domain
        self.children = []
        self.relative_time_range = relative_time_range
        if self.relative_time_range is not None:
            if isinstance(self.relative_time_range, RelativeTimeRangeFilter):
                self.relative_time_range = [self.relative_time_range]
            for rtr in self.relative_time_range:
                if rtr.anchor_phenotype is not None:
                    self.children.append(rtr.anchor_phenotype)
        super(DeathPhenotype, self).__init__(name=name, **kwargs)

    def _execute(self, tables: Dict[str, Table]) -> PhenotypeTable:
        person_table = tables[self.domain]
        person_table = person_table.mutate(EVENT_DATE=person_table.DATE_OF_DEATH)
        assert is_phenex_person_table(person_table)

        death_table = person_table.filter(person_table.DATE_OF_DEATH.notnull())
        if self.relative_time_range is not None:
            for rtr in self.relative_time_range:
                death_table = rtr.filter(death_table)
        death_table = death_table.mutate(VALUE=ibis.null())
        return death_table.mutate(EVENT_DATE=death_table.DATE_OF_DEATH)



from typing import Union
from datetime import date
from phenex.util.serialization.to_dict import to_dict


class Value:
    """
    The Value class is used to define threshold on values in databases. Importantly, Value's define not just numeric values but also the boundary (including or excluding the endpoint).

    Attributes:
       operator (str): The comparison operator, one of '>', '>=', '<', '<=', '='.
        value (Union[int, float, date]): The threshold value.

    Examples:
        greater_than_zero = Value(0, '>')
    """

    def __init__(self, operator: str, value: Union[int, float, date]):
        self.operator = operator
        self.value = value
        assert operator in [
            ">",
            ">=",
            "<",
            "<=",
            "=",
        ], "Operator must be >, >=, <, <=, or ="

    def to_dict(self):
        return to_dict(self)


class GreaterThan(Value):
    def __init__(self, value: int, **kwargs):
        super(GreaterThan, self).__init__(">", value)


class GreaterThanOrEqualTo(Value):
    def __init__(self, value: int, **kwargs):
        super(GreaterThanOrEqualTo, self).__init__(">=", value)


class LessThan(Value):
    def __init__(self, value: int, **kwargs):
        super(LessThan, self).__init__("<", value)


class LessThanOrEqualTo(Value):
    def __init__(self, value: int, **kwargs):
        super(LessThanOrEqualTo, self).__init__("<=", value)


class EqualTo(Value):
    def __init__(self, value: int, **kwargs):
        super(EqualTo, self).__init__("=", value)



from typing import Union, List, Optional
from phenex.phenotypes import CodelistPhenotype
from phenex.tables import is_phenex_code_table, PHENOTYPE_TABLE_COLUMNS, PhenotypeTable
from phenex.phenotypes.functions import select_phenotype_columns
from phenex.util import create_logger

logger = create_logger(__name__)

from ibis import _


class MeasurementPhenotype(CodelistPhenotype):
    """
    # What is MeasurementPhenotype for?
    The MeasurementPhenotype is for manipulating numerical data found in RWD data sources e.g. laboratory or observation results. These tables often contain numerical values (height, weight, blood pressure, lab results). As an event-based table, each row records a single measurement value for a single patient with a date. All numerical values are in a 'value' column. A medical code indicates the type of numerical measurement and the units of measurement are in an additional column.

    MeasurementPhenotype is a subclass of CodelistPhenotype, inheriting all of its functionality to identify patients by single or sets of medical codes (e.g. 'test type') within a specified time period. It can also :

    - identify patients with a measurement value within a value range and
    - return a measurement value, either all measurements values within filter
      criteria or perform simple aggregations (mean, median, max, min).

    # Example data:

    | PersonID    |   MedicalCode   |   EventDate   |   Value    | Unit|
    |-------------|-----------------|---------------|------------|-----|
    | 1           |   HbA1c         |   2010-01-01  |   4.2      | %   |
    | 1           |   HT            |   2010-01-02  |   121      | cm  |
    | 2           |   WT            |   2010-01-01  |   130      | kg  |

    # Note on data cleaning
    In general, data cleaning operations should be performed upstream of Phenex.
    This includes:

    - unit harmonization: ideally all values should be in the same unit for a given measurement type test. There are workarounds to deal with multiple units for a single measurement, but it is not recommended.

    - removing nonsensical values: e.g. negative blood pressures, or values outside of a physiological range. While MeasurementPhenotype provides a clean_nonphysiologicals_value_filter parameter to remove such values,  is recommended to perform this operation upstream of apex.

    - dealing with duplicate entries: e.g. multiple entries for the same patient on the same day. The meaning of multiple entries on the same day may vary between data sources, so it is recommended to handle duplicate entries upstream of apex. In some EHR datasources, duplicate entries may suggest that this value is 'more accurate', as it is passed and recorded through multiple providers and systems, while in other datasources multiple entries may suggest faulty data entry.


    Parameters:
        value_aggregation (ValueAggregator): A ValueAggregator PhenEx class defining aggregation, whether over the whole defined time period (relative time range filter, date filter) using Mean, Median, Min, Max) or daily aggregation operation (DailyMin, DailyMax, DailyMedian, DailyMean) to be performed on the measurement values. This operation occurs **after** the cleaning value filter but **prior** to the primary value_filter.
        value_filter (ValueFilter): A ValueFilter to be applied to the measurement values. This filter is applied **after** the clean_nonphysiologicals_value_filter and value_aggregation. This filter is used to identify patients with a measurement value within a value range. If not specified, no value filter is applied. For example, to identify patients with a 1. systolic blood pressure above 120 mmHg, the value_filter would be set to ValueFilter(min_value=GreaterThan(120)), 2. systolic blood pressure above 120 mmHg but below 140 mmHg, the value_filter would be set to ValueFilter(min_value=GreaterThan(120), max_value=LessThan(140)).
        further_value_filter_phenotype (str): If the input to the current MeasurementPhenotype is the output of a previous MeasurementPhenotype, set this parameter to the previous MeasurementPhenotype.
        clean_nonphysiologicals_value_filter (str): A value filter to be applied **prior** to any filtering or aggregation. This should be used to remove nonsensical values e.g. negative blood pressures, or values outside of a physiological range that are certain to be due to measurement error. Ideally, such cleaing steps should performed upstream of apex, but have been provided due to realization of practical necessity.
        clean_null_values (str): A boolean indicating whether to remove null values from the measurement table. If set to True, null values are removed prior to value filtering. If set to False, null values are not removed. If not specified, null values are removed (default is true)
        return_date (str): Specifies whether to return the 'first', 'last', or 'nearest' event date. Default is 'first'.

    """

    def __init__(
        self,
        value_filter: Optional["ValueFilter"] = None,
        clean_nonphysiologicals_value_filter: Optional["ValueFilter"] = None,
        clean_null_values: Optional[bool] = True,
        value_aggregation: Optional[
            Union[
                "ValueAggregator",
                "DailyMedian",
                "DailyMean",
                "DailyMin",
                "DailyMax",
                "DailyValueAggregator",
                "Min",
                "Max",
                "Mean",
            ]
        ] = None,
        further_value_filter_phenotype: Optional["MeasurementPhenotype"] = None,
        **kwargs,
    ):
        # Default value of return_date in codelist_phenotype is 'first'. This is not helpful behavior for measurementphenotype as we will perform further operations that require all values. For example, if we want the mean of all values in the post index period, setting return_date = 'first' will return only the values on the first day
        if "return_date" not in kwargs:
            kwargs["return_date"] = "all"
        super(MeasurementPhenotype, self).__init__(
            **kwargs,
        )
        self.clean_nonphysiologicals_value_filter = clean_nonphysiologicals_value_filter
        self.clean_null_values = clean_null_values
        self.value_filter = value_filter
        self.value_aggregation = value_aggregation
        self.further_value_filter_phenotype = further_value_filter_phenotype

        if self.return_date != "all":
            if self.value_aggregation.__class__.__name__ in [
                "Mean",
                "Median",
                "Max",
                "Min",
            ]:
                raise ValueError(
                    f"{self.name}: you have selected an aggregation of the entire time period while selecting a single date selection of {self.return_date}. Select a daily aggregator (DailyMean, DailyMedian, DailyMin, DailyMax) if selecting a specific return date."
                )

        if self.further_value_filter_phenotype is not None:
            self.children.append(self.further_value_filter_phenotype)

    def _execute(self, tables) -> PhenotypeTable:
        # perform codelist filtering
        # perform nonphysiological value filtering
        # perform value aggreation
        # perform value filter
        # perform value and dateaggregation
        code_table = tables[self.domain]
        code_table = self._perform_codelist_filtering(code_table)
        code_table = self._perform_categorical_filtering(code_table, tables)
        code_table = self._perform_null_value_filtering(code_table)
        code_table = self._perform_nonphysiological_value_filtering(code_table)
        code_table = self._perform_time_filtering(code_table)
        code_table = self._perform_date_selection(code_table, reduce=False)
        code_table = self._perform_value_aggregation(code_table)
        code_table = self._perform_value_filtering(code_table)
        return select_phenotype_columns(code_table)

    def _perform_null_value_filtering(self, code_table):
        if self.clean_null_values and self.value_filter:
            logger.debug(f"Applying null filtering for {self.name}")
            code_table = code_table[code_table[self.value_filter.column_name].notnull()]
        return code_table

    def _perform_nonphysiological_value_filtering(self, code_table):
        if self.clean_nonphysiologicals_value_filter is not None:
            code_table = self.clean_nonphysiologicals_value_filter.filter(code_table)
        return code_table

    def _perform_value_aggregation(self, code_table):
        if self.value_aggregation is not None:
            code_table = self.value_aggregation.aggregate(code_table)
        return code_table

    def _perform_value_filtering(self, code_table):
        if self.value_filter is not None:
            code_table = self.value_filter.filter(code_table)
        return code_table



from dataclasses import dataclass
from typing import Optional, Union

from phenex.codelists import Codelist
from phenex.phenotypes import (
    CodelistPhenotype,
    DeathPhenotype,
    LogicPhenotype,
    MeasurementPhenotype,
    MeasurementChangePhenotype,
)

from phenex.filters import (
    RelativeTimeRangeFilter,
    GreaterThanOrEqualTo,
    LessThanOrEqualTo,
    CategoricalFilter,
)

from phenex.util import create_logger

logger = create_logger(__name__)


@dataclass
class ISTHBleedComponents:
    """
    Database specific components of ISTH Major Bleed. These include :


    | Type                | Number                     |
    |---------------------|----------------------------|
    | codelists           | 4                          |
    | categorical filters | 4 required, 1 optional     |
    | domains             | 3                          |

    """

    critical_organ_bleed_codelist: Codelist
    overt_bleed_codelist: Codelist
    possible_bleed_codelist: Codelist
    transfusion_codelist: Codelist

    inpatient: CategoricalFilter
    outpatient: CategoricalFilter
    primary_diagnosis: CategoricalFilter
    secondary_diagnosis: CategoricalFilter
    diagnosis_of: CategoricalFilter = None

    diagnosis_code_domain: str = "CONDITION_OCCURRENCE_SOURCE"
    procedure_code_domain: str = "PROCEDURE_OCCURRENCE_SOURCE"
    measurement_code_domain: str = None
    death_domain: str = "PERSON"
    hemoglobin_codelist: Optional[Codelist] = None


def ISTHMajorBleedPhenotype(
    components: ISTHBleedComponents,
    name: Optional[str] = "isth_major_bleed",
    return_date: str = "first",
    relative_time_range: Optional[RelativeTimeRangeFilter] = None,
) -> LogicPhenotype:
    """
    Operational definition for ISTH Major Bleed as defined in [*Identification of International Society on Thrombosis and Haemostasis major and clinically relevant non-major bleed events from electronic health records: a novel algorithm to enhance data utilisation from real-world sources*,  Hartenstein et. al](https://pmc.ncbi.nlm.nih.gov/articles/PMC10898215/).

    This is a database agnostic implementation. Database specific components are specified by various ISTHBleedComponents.


    Parameters:
        components: Database specific definitions of codelists, categorical filters, and domains. See documentation for ISTHBleedComponents for more details.
        name: Optional override of default name 'isth_major_bleed'.
        return_date: Specify whether to return the date of the first, last or all bleed events
        relative_time_range: Optional specificiation of a relative time range in which to observe bleeds. For example, 'any_time_post_index' could be constructed and passed to the ISTH bleed.
    """
    critical_organ_bleed = CriticalOrganBleedPhenotype(
        components=components,
        relative_time_range=relative_time_range,
        return_date=return_date,
    )
    symptomatic_bleed = SymptomaticBleedPhenotype(
        components=components,
        relative_time_range=relative_time_range,
        return_date=return_date,
    )
    fatal_bleed = FatalBleedPhenotype(
        components=components,
        relative_time_range=relative_time_range,
        return_date=return_date,
    )
    return LogicPhenotype(
        name=name,
        expression=critical_organ_bleed | symptomatic_bleed | fatal_bleed,
        return_date=return_date,
    )


def CriticalOrganBleedPhenotype(
    components: ISTHBleedComponents,
    return_date: str = "first",
    relative_time_range: Optional[RelativeTimeRangeFilter] = None,
    name: Optional[str] = "isth_critical_organ_bleed",
) -> CodelistPhenotype:
    """
    Operational definition for critical organ bleed subcomponent of ISTH Major Bleed as defined in [*Identification of International Society on Thrombosis and Haemostasis major and clinically relevant non-major bleed events from electronic health records: a novel algorithm to enhance data utilisation from real-world sources*,  Hartenstein et. al](https://pmc.ncbi.nlm.nih.gov/articles/PMC10898215/).

    In brief, critical organ bleeds are any bleed diagnostic event from the codelist 'critical organ bleed' occurring in the inpatient setting AND (primary OR secondary) diagnosis position.

    This is a database agnostic implementation. Database specific components are specified by various ISTHBleedComponents.

    Parameters:
        components: Database specific definitions of codelists, categorical filters, and domains. See documentation for ISTHBleedComponents for more details.
        name: Optional override of default name 'isth_major_bleed'.
        return_date: Specify whether to return the date of the first, last or all bleed events
        relative_time_range: Optional specificiation of a relative time range in which to observe bleeds. For example, 'any_time_post_index' could be constructed and passed to the ISTH bleed.
    """
    # create required categorical filters; critical organ bleed occurs only in the inpatient position and can be either primary or secondary diagnosis position
    categorical_filters = components.inpatient & (
        components.primary_diagnosis | components.secondary_diagnosis
    )
    categorical_filters = add_diagnosis_of_filter(categorical_filters, components)

    return CodelistPhenotype(
        name=name,
        domain=components.diagnosis_code_domain,
        codelist=components.critical_organ_bleed_codelist,
        categorical_filter=categorical_filters,
        relative_time_range=relative_time_range,
        return_date=component_return_date(relative_time_range),
    )


def SymptomaticBleedPhenotype(
    components: ISTHBleedComponents,
    return_date: str = "first",
    relative_time_range: Optional[RelativeTimeRangeFilter] = None,
    name: Optional[str] = "isth_overt_bleed",
) -> "BleedVerificationPhenotype":
    """
    Operational definition for symptomatic bleeds as defined in [*Identification of International Society on Thrombosis and Haemostasis major and clinically relevant non-major bleed events from electronic health records: a novel algorithm to enhance data utilisation from real-world sources*,  Hartenstein et. al](https://pmc.ncbi.nlm.nih.gov/articles/PMC10898215/).

    In brief, symptomatic bleeds are any bleed diagnostic events from the codelist 'overt_bleed' occurring in the inpatient setting AND the primary diagnosis position AND having a blood transfusion procedure event or hemoglobin drop lab event within +/- 2 days of diagnosis event.

    This is a database agnostic implementation. Database specific components are specified by various ISTHBleedComponents.

    Parameters:
        components: Database specific definitions of codelists, categorical filters, and domains. See documentation for ISTHBleedComponents for more details.
        name: Optional override of default name 'isth_major_bleed'.
        return_date: Specify whether to return the date of the first, last or all bleed events
        relative_time_range: Optional specificiation of a relative time range in which to observe bleeds. For example, 'any_time_post_index' could be constructed and passed to the ISTH bleed.
    """
    # create required categorical filters; critical organ bleed occurs only in the inpatient position and can be the primary diagnosis position
    categorical_filters = components.inpatient & components.primary_diagnosis

    categorical_filters = add_diagnosis_of_filter(categorical_filters, components)

    overt_bleed = CodelistPhenotype(
        name=name,
        domain=components.diagnosis_code_domain,
        codelist=components.overt_bleed_codelist,
        categorical_filter=categorical_filters,
        relative_time_range=relative_time_range,
        return_date=component_return_date(relative_time_range),
    )

    return BleedVerificationPhenotype(
        anchor_phenotype=overt_bleed,
        components=components,
        relative_time_range=relative_time_range,
    )


def FatalBleedPhenotype(
    components: ISTHBleedComponents,
    return_date: str = "first",
    relative_time_range: Optional[RelativeTimeRangeFilter] = None,
    name: Optional[str] = "isth_fatal_bleed",
) -> CodelistPhenotype:
    """
    Operational definition for fatal bleed subcomponent of ISTH Major Bleed as defined in [*Identification of International Society on Thrombosis and Haemostasis major and clinically relevant non-major bleed events from electronic health records: a novel algorithm to enhance data utilisation from real-world sources*,  Hartenstein et. al](https://pmc.ncbi.nlm.nih.gov/articles/PMC10898215/).

    In brief, fatal bleeds are bleed diagnostic events from the codelist 'critical_organ_bleed' that occur with 45 days prior to a death event.

    This is a database agnostic implementation. Database specific components are specified by various ISTHBleedComponents.

    Parameters:
        components: Database specific definitions of codelists, categorical filters, and domains. See documentation for ISTHBleedComponents for more details.
        name: Optional override of default name 'isth_major_bleed'.
        return_date: Specify whether to return the date of the first, last or all bleed events
        relative_time_range: Optional specificiation of a relative time range in which to observe bleeds. For example, 'any_time_post_index' could be constructed and passed to the ISTH bleed.
    """
    death = DeathPhenotype(
        relative_time_range=relative_time_range, domain=components.death_domain
    )

    relative_45_days_prior = RelativeTimeRangeFilter(
        when="before",
        min_days=GreaterThanOrEqualTo(0),
        max_days=LessThanOrEqualTo(45),
        anchor_phenotype=death,
    )

    # create required categorical filters; critical organ bleed already covers the inpatient conditions. we must only look for outpatient and primary diagnosis positions
    categorical_filters = components.outpatient & components.primary_diagnosis
    categorical_filters = add_diagnosis_of_filter(
        categorical_filters, components=components
    )

    if relative_time_range is None:
        full_relative_time_range = relative_45_days_prior
    else:
        full_relative_time_range = [relative_45_days_prior, relative_time_range]
    return CodelistPhenotype(
        name=name,
        domain=components.diagnosis_code_domain,
        codelist=components.critical_organ_bleed_codelist,
        categorical_filter=categorical_filters,
        relative_time_range=full_relative_time_range,
        return_date=component_return_date(relative_time_range),
    )


def BleedVerificationPhenotype(
    anchor_phenotype,
    name: Optional[str] = None,
    components: ISTHBleedComponents = None,
    relative_time_range: Optional[RelativeTimeRangeFilter] = None,
    transfusion_codelist: Optional["Codelist"] = None,
    procedure_code_domain: Optional[str] = None,
    hemoglobin_codelist: Optional["Codelist"] = None,
    measurement_code_domain: Optional[str] = None,
) -> CodelistPhenotype:
    """
    Operational definition for 'bleed verification' of ISTH Major Bleed as defined in [*Identification of International Society on Thrombosis and Haemostasis major and clinically relevant non-major bleed events from electronic health records: a novel algorithm to enhance data utilisation from real-world sources*,  Hartenstein et. al](https://pmc.ncbi.nlm.nih.gov/articles/PMC10898215/).

    In brief, bleed verification is

    This is a database agnostic implementation. Database specific components are specified by various ISTHBleedComponents.

    Parameters:
        components: Database specific definitions of codelists, categorical filters, and domains. See documentation for ISTHBleedComponents for more details.
        name: Optional override of default name 'isth_major_bleed'.
        anchor_phenotype: Specify whether to return the date of the first, last or all bleed events
        relative_time_range: Optional specificiation of a relative time range in which to observe bleeds. For example, 'any_time_post_index' could be constructed and passed to the ISTH bleed.
    """
    if components is not None:
        transfusion_codelist = components.transfusion_codelist
        procedure_code_domain = components.procedure_code_domain
        hemoglobin_codelist = components.hemoglobin_codelist
        measurement_code_domain = components.measurement_code_domain
    if name is None:
        name = "bleed_verification_for_" + anchor_phenotype.name

    within_two_days = RelativeTimeRangeFilter(
        when="after",
        min_days=GreaterThanOrEqualTo(-2),
        max_days=LessThanOrEqualTo(2),
        anchor_phenotype=anchor_phenotype,
    )

    if relative_time_range is not None:
        composite_relative_time_range = [relative_time_range, within_two_days]

    transfusion = BloodTransfusionPhenotype(
        anchor_phenotype=anchor_phenotype,
        name=f"{name}_blood_transfusion",
        components=components,
        relative_time_range=composite_relative_time_range,
        transfusion_codelist=transfusion_codelist,
        procedure_code_domain=procedure_code_domain,
    )

    if hemoglobin_codelist is None and measurement_code_domain is None:
        logger.info(
            "ISTH Major bleeding verification : no hemoglobing measurement codes or measurement table provided. Using procedure codes for blood transfusion only."
        )
        return transfusion
    hb_drop = HbDropPhenotype(
        anchor_phenotype=anchor_phenotype,
        name=f"{name}_hb_drop",
        components=components,
        relative_time_range=composite_relative_time_range,
        hemoglobin_codelist=hemoglobin_codelist,
        measurement_code_domain=measurement_code_domain,
    )
    logger.info(
        "ISTH Major bleeding verification : using both transfusion and hemoglobin information."
    )

    return LogicPhenotype(
        name=name,
        expression=transfusion | hb_drop,
        return_date=component_return_date(relative_time_range),
    )


def BloodTransfusionPhenotype(
    anchor_phenotype,
    name: Optional[str] = None,
    components: ISTHBleedComponents = None,
    relative_time_range: Optional[RelativeTimeRangeFilter] = None,
    transfusion_codelist: Optional["Codelist"] = None,
    procedure_code_domain: Optional[str] = None,
) -> CodelistPhenotype:
    """
    Operational definition for 'bleed verification' of ISTH Major Bleed as defined in [*Identification of International Society on Thrombosis and Haemostasis major and clinically relevant non-major bleed events from electronic health records: a novel algorithm to enhance data utilisation from real-world sources*,  Hartenstein et. al](https://pmc.ncbi.nlm.nih.gov/articles/PMC10898215/).

    In brief, bleed verification is

    This is a database agnostic implementation. Database specific components are specified by various ISTHBleedComponents.

    Parameters:
        components: Database specific definitions of codelists, categorical filters, and domains. See documentation for ISTHBleedComponents for more details.
        name: Optional override of default name 'isth_major_bleed'.
        anchor_phenotype: Specify whether to return the date of the first, last or all bleed events
        relative_time_range: Optional specificiation of a relative time range in which to observe bleeds. For example, 'any_time_post_index' could be constructed and passed to the ISTH bleed.
    """
    if components is not None:
        transfusion_codelist = components.transfusion_codelist
        procedure_code_domain = components.procedure_code_domain
    if name is None:
        name = "transfusion_two_days_within_" + anchor_phenotype.name

    transfusion = CodelistPhenotype(
        name=name,
        codelist=transfusion_codelist,
        domain=procedure_code_domain,
        relative_time_range=relative_time_range,
        return_date="all",
    )
    return transfusion


def HbDropPhenotype(
    anchor_phenotype,
    name: Optional[str] = None,
    components: ISTHBleedComponents = None,
    relative_time_range: Optional[RelativeTimeRangeFilter] = None,
    hemoglobin_codelist: Optional["Codelist"] = None,
    measurement_code_domain: Optional[str] = None,
) -> CodelistPhenotype:
    if components is not None:
        hemoglobin_codelist = components.hemoglobin_codelist
        measurement_code_domain = components.measurement_code_domain
    if name is None:
        name = "hb_drop_two_days_within_" + anchor_phenotype.name

    hemoglobin = MeasurementPhenotype(
        name=f"{name}_hemoglobin",
        codelist=hemoglobin_codelist,
        domain=measurement_code_domain,
        relative_time_range=relative_time_range,
    )

    hemoglobin_drop = MeasurementChangePhenotype(
        name=name,
        phenotype=hemoglobin,
        min_change=GreaterThanOrEqualTo(2),
        max_days_between=LessThanOrEqualTo(2),
        direction="decrease",
        component_date_select="second",
        return_date="all",
    )

    return hemoglobin_drop


def add_diagnosis_of_filter(categorical_filters, components):
    # if a 'diagnosis_of' (as opposed to 'history_of', etc) categorical filter is provided, add it to the list of conditions
    if components.diagnosis_of is not None:
        categorical_filters = categorical_filters & components.diagnosis_of
    return categorical_filters


def component_return_date(relative_time_range):
    if relative_time_range is None:
        return "first"
    if relative_time_range.when == "after":
        return "first"
    return "last"



from ibis.expr.types.relations import Table
from phenex.tables import PhenexTable
from typing import Optional, Dict
from phenex.util.serialization.to_dict import to_dict


class Filter:
    """
    Filters operate on single tables and return these tables with rows removed. Filters are generally used within a Phenotype as a subquery. Filters know about their dependencies but cannot trigger recursive execution. Fitlers can add columns but may not remove columns. All classes in the filters module should subclass this class. This class is not intended to be used directly, but should be subclassed to create useful filters. Subclasses must implement the _filter method.

    Methods:
        filter: Filters the given table according to the rules of the Filter.

        autojoin_filter: Like filter, but magically joins needed tables in order to filter table, for instance, when filtering a CONDITION_OCCURRENCE-like table based on contextual information in a VISIT-like table.
    """

    def __init__(self):
        pass

    def filter(self, table: PhenexTable) -> PhenexTable:
        """
        Filters the given table according to the rules of the Filter.

        Args:
            table (PhenexTable): The table to be filtered.

        Returns:
            PhenexTable: The filtered table. The returned table has the exact same schema as the input table but has rows removed.
        """
        input_columns = table.columns
        filtered_table = self._filter(table)
        if not set(input_columns) <= set(filtered_table.columns):
            raise ValueError(f"Filter must not remove columns.")

        filtered_table = filtered_table.select(input_columns)
        if isinstance(table, PhenexTable):
            return type(table)(filtered_table)
        else:
            return filtered_table

    def _filter(self, table: Table) -> Table:
        """
        Performs the operations required to filter the table.
        """
        raise NotImplementedError()

    def autojoin_filter(
        self, table: "PhenexTable", tables: Optional[Dict[str, "PhenexTable"]] = None
    ) -> "PhenexTable":
        raise NotImplementedError()

    def __and__(self, other):
        return AndFilter(self, other)

    def __or__(self, other):
        return OrFilter(self, other)

    def __invert__(self):
        return NotFilter(self)

    def to_dict(self):
        return to_dict(self)


class AndFilter(Filter):
    """
    Combines two filters using logical AND.
    """

    def __init__(self, filter1, filter2):
        self.filter1 = filter1
        self.filter2 = filter2

    def filter(self, table: Table) -> Table:
        table = self.filter1.filter(table)
        return self.filter2.filter(table)

    def autojoin_filter(self, table: "PhenexTable", tables: dict = None):
        table = self.filter1.autojoin_filter(table, tables)
        return self.filter2.autojoin_filter(table, tables)


class OrFilter(Filter):
    """
    Combines two filters using logical OR.
    """

    def __init__(self, filter1, filter2):
        self.filter1 = filter1
        self.filter2 = filter2

    def filter(self, table: Table) -> Table:
        table1 = self.filter1.filter(table).table
        table2 = self.filter2.filter(table).table
        return type(table)(table1.union(table2).distinct())

    def autojoin_filter(self, table: "PhenexTable", tables: dict = None):
        table1 = self.filter1.autojoin_filter(table, tables).table
        table2 = self.filter2.autojoin_filter(table, tables).table
        return type(table)(table1.union(table2, distinct=True))


class NotFilter(Filter):
    """
    Negates a filter.
    """

    def __init__(self, filter):
        self.filter = filter

    def filter(self, table: Table) -> Table:
        filtered_table = self.filter.filter(table).table
        return type(table)(table.difference(filtered_table))

    def autojoin_filter(self, table: "PhenexTable", tables: dict = None):
        filtered_table = self.filter.autojoin_filter(table, tables).table
        return type(table)(table.difference(filtered_table))



import inspect
from datetime import date, datetime


def to_dict(obj) -> dict:
    """
    This Function is used to serialize PhenEx objects into a format that can be easily stored or transmitted.
    Given a PhenEx object, it will return a dictionary with the class name and its parameters.
    This method runs recursively on all parameters of a PhenEx object.

    PhenEx classes generally have their own to_dict method. Those class methods call this method.
    Generally users are recommended to use the class methods instead of this method.
    This method is used internally by the class methods and by the PhenEx JSON-like serialization methods.
    """

    init_params = get_phenex_init_params(obj.__class__)
    _dict = {"class_name": obj.__class__.__name__}
    for param in init_params:
        if param != "self":
            value = getattr(obj, param, None)
            if isinstance(value, list):
                _dict[param] = [
                    (
                        item.to_dict()
                        if hasattr(item, "to_dict") and callable(item.to_dict)
                        else item
                    )
                    for item in value
                ]
            elif hasattr(value, "to_dict") and callable(value.to_dict):
                _dict[param] = value.to_dict()
            elif isinstance(value, (date, datetime)):
                _dict[param] = {"__datetime__": value.isoformat()}
            else:
                _dict[param] = value

    _dict.pop("kwargs", None)

    return _dict


def get_phenex_init_params(cls) -> dict:
    """
    Get all initialization parameters used to construct a PhenEx class.
    """
    params = {}
    if cls.__module__.startswith("phenex"):
        for base in cls.__bases__:
            params.update(get_phenex_init_params(base))
        params.update(inspect.signature(cls.__init__).parameters)
    return params



from phenex.tables import PhenexPersonTable, CodeTable, PhenexObservationPeriodTable


class PersonTableForTests(PhenexPersonTable):
    NAME_TABLE = "PATIENT"
    DEFAULT_MAPPING = {
        "PERSON_ID": "PATID",
        "YEAR_OF_BIRTH": "YOB",
        "SEX": "GENDER",
    }
    JOIN_KEYS = {
        "ConditionOccurenceTableForTests": ["PERSON_ID"],
        "TestVisitDetailTable": ["PERSON_ID"],
    }


class ConditionOccurenceTableForTests(CodeTable):
    NAME_TABLE = "OBSERVATION"
    JOIN_KEYS = {
        "PersonTableForTests": ["PATID"],
    }
    DEFAULT_MAPPING = {
        "PERSON_ID": "PATID",
        "EVENT_DATE": "OBSDATE",
        "CODE": "MEDCODEID",
    }


class DrugExposureTableForTests(CodeTable):
    NAME_TABLE = "DRUGISSUE"
    JOIN_KEYS = {
        "PersonTableForTests": ["PATID"],
    }
    DEFAULT_MAPPING = {
        "PERSON_ID": "PATID",
        "EVENT_DATE": "ISSUEDATE",
        "CODE": "PRODCODEID",
    }


class ObservationPeriodTableForTests(PhenexObservationPeriodTable):
    NAME_TABLE = "PATIENT"
    JOIN_KEYS = {"PersonTableForTests": ["PATID"]}
    DEFAULT_MAPPING = {
        "PERSON_ID": "PATID",
        "START_DATE": "REGSTARTDATE",
        "END_DATE": "REGENDDATE",
    }



from typing import List, Optional, Union
from phenex.phenotypes.categorical_phenotype import CategoricalPhenotype
from phenex.filters import CategoricalFilter


class SexPhenotype(CategoricalPhenotype):
    """
    SexPhenotype represents a sex-based phenotype. It returns the sex of individuals in the VALUE column and optionally filters based on identified sex. DATE is not defined for SexPhenotype.

    Parameters:
        name: Name of the phenotype, default is 'sex'.
        domain: Domain of the phenotype, default is 'PERSON'.
        allowed_values: List of allowed values for the categorical variable.
        column_name: Name of the column containing the required categorical variable. Default is 'SEX'.

    Examples:

    Example: Return the recorded sex of all patients.
    ```python
    from phenex.phenotypes import SexPhenotype
    sex = SexPhenotype()
    ```

    Example: Extract all male patients from the database.
    ```python
    from phenex.phenotypes import SexPhenotype
    sex = SexPhenotype(
        allowed_values=['M'],
        column_name='GENDER_SOURCE_VALUE'
        )
    ```
    """

    def __init__(
        self,
        name: str = "SEX",
        domain: str = "PERSON",
        categorical_filter: "CategoricalFilter" = None,
        **kwargs
    ):
        if categorical_filter is None:
            categorical_filter = CategoricalFilter(column_name="SEX")
        else:
            if categorical_filter.column_name is None:
                categorical_filter.column_name = "SEX"

        super(SexPhenotype, self).__init__(
            name=name, domain=domain, categorical_filter=categorical_filter, **kwargs
        )



import os
from typing import Dict, List, Union, Optional
import pandas as pd
import warnings
from phenex.util.serialization.to_dict import to_dict


class Codelist:
    """
    Codelist is a class that allows us to conveniently work with medical codes used in RWD analyses. A Codelist represents a (single) specific medical concept, such as 'atrial fibrillation' or 'myocardial infarction'. A Codelist is associated with a set of medical codes from one or multiple source vocabularies (such as ICD10CM or CPT); we call these vocabularies 'code types'. Code type is important, as there are no assurances that codes from different vocabularies (different code types) do not overlap. It is therefore highly recommended to always specify the code type when using a codelist.

    Codelist is a simple class that stores the codelist as a dictionary. The dictionary is keyed by code type and the value is a list of codes. Codelist also has various convenience methods such as read from excel, csv or yaml files, and export to excel files.

    Fuzzy codelists allow the use of '%' as a wildcard character in codes. This can be useful when you want to match a range of codes that share a common prefix. For example, 'I48.%' will match any code that starts with 'I48.'. Multiple fuzzy matches can be passed just like ordinary codes in a list.

    If a codelist contains more than 100 fuzzy codes, a warning will be issued as performance may suffer significantly.

    Parameters:
        name: Descriptive name of codelist
        codelist: User can enter codelists as either a string, a list of strings or a dictionary keyed by code type. In first two cases, the class will convert the input to a dictionary with a single key None. All consumers of the Codelist instance can then assume the codelist in that format.
        use_code_type: User can define whether code type should be used or not.
        remove_punctuation: User can define whether punctuation should be removed from codes or not.

    Methods:
        from_yaml: Load a codelist from a YAML file.
        from_excel: Load a codelist from an Excel file.
        from_csv: Load a codelist from a CSV file.

    File Formats:
        YAML:
        The YAML file should contain a dictionary where the keys are code types
        (e.g., "ICD-9", "ICD-10") and the values are lists of codes for each type.

        Example:
        ```yaml
        ICD-9:
          - "427.31"  # Atrial fibrillation
        ICD-10:
          - "I48.0"   # Paroxysmal atrial fibrillation
          - "I48.1"   # Persistent atrial fibrillation
          - "I48.2"   # Chronic atrial fibrillation
          - "I48.91"  # Unspecified atrial fibrillation
        ```

        Excel:
        The Excel file should contain a minimum of two columns for code and code_type. If multiple codelists exist in the same table, an additional column for codelist names is required.

        Example (Single codelist):
        ```markdown
        | code_type | code   |
        |-----------|--------|
        | ICD-9     | 427.31 |
        | ICD-10    | I48.0  |
        | ICD-10    | I48.1  |
        | ICD-10    | I48.2  |
        | ICD-10    | I48.91 |
        ```

        Example (Multiple codelists):
        ```markdown
        | code_type | code   | codelist           |
        |-----------|--------|--------------------|
        | ICD-9     | 427.31 | atrial_fibrillation|
        | ICD-10    | I48.0  | atrial_fibrillation|
        | ICD-10    | I48.1  | atrial_fibrillation|
        | ICD-10    | I48.2  | atrial_fibrillation|
        | ICD-10    | I48.91 | atrial_fibrillation|
        ```

        CSV:
        The CSV file should follow the same format as the Excel file, with columns for code, code_type, and optionally codelist names.

    Example:
    ```python
    # Initialize with a list
    cl = Codelist(
        ['x', 'y', 'z'],
        'mycodelist'
        )
    print(cl.codelist)
    {None: ['x', 'y', 'z']}
    ```

    Example:
    ```python
    # Initialize with string
    cl = Codelist(
        'SBP'
        )
    print(cl.codelist)
    {None: ['SBP']}
    ```

    Example:
    ```python
    # Initialize with a dictionary
    >> atrial_fibrillation_icd_codes = {
        "ICD-9": [
            "427.31"  # Atrial fibrillation
        ],
        "ICD-10": [
            "I48.0",  # Paroxysmal atrial fibrillation
            "I48.1",  # Persistent atrial fibrillation
            "I48.2",  # Chronic atrial fibrillation
            "I48.91", # Unspecified atrial fibrillation
        ]
    }
    cl = Codelist(
        atrial_fibrillation_icd_codes,
        'atrial_fibrillation',
    )
    print(cl.codelist)
    {
        "ICD-9": [
            "427.31"  # Atrial fibrillation
        ],
        "ICD-10": [
            "I48.0",  # Paroxysmal atrial fibrillation
            "I48.1",  # Persistent atrial fibrillation
            "I48.2",  # Chronic atrial fibrillation
            "I48.91", # Unspecified atrial fibrillation
        ]
    }
    ```

    ```python
    # Initialize with a fuzzy codelist
    anemia = Codelist(
        {'ICD10CM': ['D55%', 'D56%', 'D57%', 'D58%', 'D59%', 'D60%']},
        {'ICD9CM': ['284%', '285%', '282%']},
        'fuzzy_codelist'
    )
    ```
    """

    def __init__(
        self,
        codelist: Union[str, List, Dict[str, List]],
        name: Optional[str] = None,
        use_code_type: Optional[bool] = True,
        remove_punctuation: Optional[bool] = False,
    ) -> None:
        self.name = name

        if isinstance(codelist, dict):
            self.codelist = codelist
        elif isinstance(codelist, list):
            self.codelist = {None: codelist}
        elif isinstance(codelist, str):
            if name is None:
                self.name = codelist
            self.codelist = {None: [codelist]}
        else:
            raise TypeError("Input codelist must be a dictionary, list, or string.")

        if list(self.codelist.keys()) == [None]:
            self.use_code_type = False
        else:
            self.use_code_type = use_code_type

        self.remove_punctuation = remove_punctuation

        self.fuzzy_match = False
        for code_type, codelist in self.codelist.items():
            if any(["%" in str(code) for code in codelist]):
                self.fuzzy_match = True
                if len(codelist) > 100:
                    warnings.warn(
                        f"Detected fuzzy codelist match with > 100 regex's for code type {code_type}. Performance may suffer significantly."
                    )

        self._resolved_codelist = None

    def copy(
        self,
        name: Optional[str] = None,
        use_code_type: bool = True,
        remove_punctuation: bool = False,
        rename_code_type: dict = None,
    ) -> "Codelist":
        """
        Codelist's are immutable. If you want to update how codelists are resolved, make a copy of the given codelist changing the resolution parameters.

        Parameters:
            name: Name for newly created code list if different from the old one.
            use_code_type: If False, merge all the code lists into one with None as the key.
            remove_punctuation: If True, remove '.' from all codes.
            rename_code_type: Dictionary defining code types that should be renamed. For example, if the original code type is 'ICD-10-CM', but it is 'ICD10' in the database, we must rename the code type. This keyword argument is a dictionary with keys being the current code type and the value being the desired code type. Code types not included in the mapping are left unchanged.

        Returns:
            Codelist instance with the updated resolution options.
        """
        _codelist = self.codelist.copy()
        if rename_code_type is not None and isinstance(rename_code_type, dict):
            for current, renamed in rename_code_type.items():
                if _codelist.get(current) is not None:
                    _codelist[renamed] = _codelist[current]
                    del _codelist[current]

        return Codelist(
            _codelist,
            name=name or self.name,
            use_code_type=use_code_type,
            remove_punctuation=remove_punctuation,
        )

    @property
    def resolved_codelist(self):
        """
        Retrieve the actual codelists used for filtering after processing for punctuation and code type options (see __init__()).
        """
        if self._resolved_codelist is None:
            resolved_codelist = {}

            for code_type, codes in self.codelist.items():
                if self.remove_punctuation:
                    codes = [code.replace(".", "") for code in codes]
                if self.use_code_type:
                    resolved_codelist[code_type] = codes
                else:
                    if None not in resolved_codelist:
                        resolved_codelist[None] = []
                    resolved_codelist[None] = list(
                        set(resolved_codelist[None]) | set(codes)
                    )
            self._resolved_codelist = resolved_codelist

        return self._resolved_codelist

    @classmethod
    def from_yaml(cls, path: str) -> "Codelist":
        """
        Load a codelist from a yaml file.

        The YAML file should contain a dictionary where the keys are code types
        (e.g., "ICD-9", "ICD-10") and the values are lists of codes for each type.

        Example:
        ```yaml
        ICD-9:
          - "427.31"  # Atrial fibrillation
        ICD-10:
          - "I48.0"   # Paroxysmal atrial fibrillation
          - "I48.1"   # Persistent atrial fibrillation
          - "I48.2"   # Chronic atrial fibrillation
          - "I48.91"  # Unspecified atrial fibrillation
        ```

        Parameters:
            path: Path to the YAML file.

        Returns:
            Codelist instance.
        """
        import yaml

        with open(path, "r") as f:
            data = yaml.safe_load(f)
        return cls(
            data, name=os.path.basename(path.replace(".yaml", "").replace(".yml", ""))
        )

    @classmethod
    def from_excel(
        cls,
        path: str,
        sheet_name: Optional[str] = None,
        codelist_name: Optional[str] = None,
        code_column: Optional[str] = "code",
        code_type_column: Optional[str] = "code_type",
        codelist_column: Optional[str] = "codelist",
    ) -> "Codelist":
        """
         Load a single codelist located in an Excel file.

         It is required that the Excel file contains a minimum of two columns for code and code_type. The actual columnnames can be specified using the code_column and code_type_column parameters.

         If multiple codelists exist in the same excel table, the codelist_column and codelist_name are required to point to the specific codelist of interest.

         It is possible to specify the sheet name if the codelist is in a specific sheet.

         1. Single table, single codelist : The table (whether an entire excel file, or a single sheet in an excel file) contains only one codelist. The table should have columns for code and code_type.

             ```markdown
             | code_type | code   |
             |-----------|--------|
             | ICD-9     | 427.31 |
             | ICD-10    | I48.0  |
             | ICD-10    | I48.1  |
             | ICD-10    | I48.2  |
             | ICD-10    | I48.91 |
             ```

        2. Single table, multiple codelists: A single table (whether an entire file, or a single sheet in an excel file) contains multiple codelists. A column for the name of each codelist is required. Use codelist_name to point to the specific codelist of interest.

             ```markdown
             | code_type | code   | codelist           |
             |-----------|--------|--------------------|
             | ICD-9     | 427.31 | atrial_fibrillation|
             | ICD-10    | I48.0  | atrial_fibrillation|
             | ICD-10    | I48.1  | atrial_fibrillation|
             | ICD-10    | I48.2  | atrial_fibrillation|
             | ICD-10    | I48.91 | atrial_fibrillation|
             ```

         Parameters:
             path: Path to the Excel file.
             sheet_name: An optional label for the sheet to read from. If defined, the codelist will be taken from that sheet. If no sheet_name is defined, the first sheet is taken.
             codelist_name: An optional name of the codelist which to extract. If defined, codelist_column must be present and the codelist_name must occur within the codelist_column.
             code_column: The name of the column containing the codes.
             code_type_column: The name of the column containing the code types.
             codelist_column: The name of the column containing the codelist names.

         Returns:
             Codelist instance.
        """
        import pandas as pd

        if sheet_name is None:
            _df = pd.read_excel(path)
        else:
            xl = pd.ExcelFile(path)
            if sheet_name not in xl.sheet_names:
                raise ValueError(
                    f"Sheet name {sheet_name} not found in the Excel file."
                )
            _df = xl.parse(sheet_name)

        if codelist_name is not None:
            # codelist name is not none, therefore we subset the table to the current codelist
            _df = _df[_df[codelist_column] == codelist_name]

        code_dict = _df.groupby(code_type_column)[code_column].apply(list).to_dict()

        if codelist_name is not None:
            name = codelist_name
        elif sheet_name is not None:
            name = sheet_name
        else:
            name = path.split(os.sep)[-1].replace(".xlsx", "")

        return cls(code_dict, name=name)

    @classmethod
    def from_csv(
        cls,
        path: str,
        codelist_name: Optional[str] = None,
        code_column: Optional[str] = "code",
        code_type_column: Optional[str] = "code_type",
        codelist_column: Optional[str] = "codelist",
    ) -> "Codelist":
        _df = pd.read_csv(path)

        if codelist_name is not None:
            # codelist name is not none, therefore we subset the table to the current codelist
            _df = _df[_df[codelist_column] == codelist_name]

        code_dict = _df.groupby(code_type_column)[code_column].apply(list).to_dict()

        if codelist_name is None:
            name = codelist_name
        else:
            name = path.split(os.sep)[-1].replace(".csv", "")

        return cls(code_dict, name=name)

    @classmethod
    def from_medconb(cls, codelist):
        """
        Converts a MedConB style Codelist into a PhenEx style codelist.

        Example:

        ```python
        from medconb_client import Client
        endpoint = "https://api.medconb.example.com/graphql/"
        token = get_token()
        client = Client(endpoint, token)

        medconb_codelist = client.get_codelist(
            codelist_id="9c4ad312-3008-4d95-9b16-6f9b21ec1ad9"
        )
        phenex_codelist = Codelist.from_medconb(medconb_codelist)
        ```
        """
        phenex_codelist = {}
        for codeset in codelist.codesets:
            phenex_codelist[codeset.ontology] = [c[0] for c in codeset.codes]
        return cls(codelist=phenex_codelist, name=codelist.name)

    def to_tuples(self) -> List[tuple]:
        """
        Convert the codelist to a list of tuples, where each tuple is of the form
        (code_type, code).
        """
        return sum(
            [[(ct, c) for c in self.codelist[ct]] for ct in self.codelist.keys()],
            [],
        )

    def __repr__(self):
        return f"""Codelist(
    name='{self.name}',
    codelist={self.codelist}
)"""

    def to_pandas(self) -> pd.DataFrame:
        """
        Export the codelist to a pandas DataFrame. The DataFrame will have three columns: code_type, code, and codelist.
        """

        _df = pd.DataFrame(self.to_tuples(), columns=["code_type", "code"])
        _df["codelist"] = self.name
        return _df

    def to_dict(self):
        return to_dict(self)

    def __add__(self, other):
        codetypes = list(set(list(self.codelist.keys()) + list(other.codelist.keys())))
        new_codelist = {}
        for codetype in codetypes:
            new_codelist[codetype] = list(
                set(self.codelist.get(codetype, []) + other.codelist.get(codetype, []))
            )
        if self.remove_punctuation != other.remove_punctuation:
            raise ValueError(
                "Cannot add codelists with non-matching remove_punctuation settings."
            )
        if self.use_code_type != other.use_code_type:
            raise ValueError(
                "Cannot add codelists with non-matching use_code_type settings."
            )

        return Codelist(
            new_codelist,
            name=f"({self.name}_union_{other.name})",
            remove_punctuation=self.remove_punctuation,
            use_code_type=self.use_code_type,
        )

    def __sub__(self, other):
        codetypes = list(self.codelist.keys())
        new_codelist = {}
        for codetype in codetypes:
            new_codelist[codetype] = [
                x
                for x in self.codelist.get(codetype, [])
                if x not in other.codelist.get(codetype, [])
            ]

        if self.remove_punctuation != other.remove_punctuation:
            raise ValueError(
                "Cannot create difference of codelists with non-matching remove_punctuation settings."
            )
        if self.use_code_type != other.use_code_type:
            raise ValueError(
                "Cannot create difference of codelists with non-matching use_code_type settings."
            )

        return Codelist(
            new_codelist,
            name=f"{self.name}_excluding_{other.name}",
            remove_punctuation=self.remove_punctuation,
            use_code_type=self.use_code_type,
        )


class LocalCSVCodelistFactory:
    """
    LocalCSVCodelistFactory allows for the creation of multiple codelists from a single CSV file. Use this class when you have a single CSV file that contains multiple codelists.

    To use, create an instance of the class and then call the `get_codelist` method with the name of the codelist you want to retrieve; this codelist name must be an entry in the name_codelist_column.
    """

    def __init__(
        self,
        path: str,
        name_code_column: str = "code",
        name_codelist_column: str = "codelist",
        name_code_type_column: str = "code_type",
    ) -> None:
        """
        Parameters:
            path: Path to the CSV file.
            name_code_column: The name of the column containing the codes.
            name_codelist_column: The name of the column containing the codelist names.
            name_code_type_column: The name of the column containing the code types.
        """
        self.path = path
        self.name_code_column = name_code_column
        self.name_codelist_column = name_codelist_column
        self.name_code_type_column = name_code_type_column
        try:
            self.df = pd.read_csv(path)
        except:
            raise ValueError("Could not read the file at the given path.")

        # Check if the required columns exist in the DataFrame
        required_columns = [
            name_code_column,
            name_codelist_column,
            name_code_type_column,
        ]
        missing_columns = [
            col for col in required_columns if col not in self.df.columns
        ]
        if missing_columns:
            raise ValueError(
                f"The following required columns are missing in the CSV: {', '.join(missing_columns)}"
            )

    def get_codelists(self) -> List[str]:
        """
        Get a list of all codelists in the supplied CSV.
        """
        return self.df[self.name_codelist_column].unique().tolist()

    def get_codelist(self, name: str) -> Codelist:
        """
        Retrieve a single codelist by name.
        """
        try:
            df_codelist = self.df[self.df[self.name_codelist_column] == name]
            code_dict = (
                df_codelist.groupby(self.name_code_type_column)[self.name_code_column]
                .apply(list)
                .to_dict()
            )
            return Codelist(name=name, codelist=code_dict)
        except:
            raise ValueError("Could not find the codelist with the given name.")


class MedConBCodelistFactory:
    """
    Retrieve Codelists for use in Phenex from MedConB.

    Example:
    ```python
    from medconb_client import Client
    endpoint = "https://api.medconb.example.com/graphql/"
    token = get_token()
    client = Client(endpoint, token)
    medconb_factory = MedConBCodelistFactory(client)

    phenex_codelist = medconb_factory.get_codelist(
        id="9c4ad312-3008-4d95-9b16-6f9b21ec1ad9"
    )
    ```
    """

    def __init__(
        self,
        medconb_client,
    ):
        self.medconb_client = medconb_client

    def get_codelist(self, id: str):
        """
        Resolve the codelist by querying the MedConB client.
        """
        medconb_codelist = self.medconb_client.get_codelist(codelist_id=id)
        return Codelist.from_medconb(medconb_codelist)

    def get_codelists(self):
        """
        Returns a list of all available codelist IDs.
        """
        return sum(
            [c.items for c in self.medconb_client.get_workspace().collections], []
        )



import os, logging


def create_logger(name):
    # Create a logger
    logger = logging.getLogger(name)
    log_level = os.environ.get("PHENEX_LOG_LEVEL", "DEBUG").upper()
    logger.setLevel(log_level)

    # Only add handler if none exist to prevent duplicate handlers
    if not logger.handlers:
        # Create console handler and set level to debug
        console_handler = logging.StreamHandler()
        console_handler.setLevel(log_level)

        # Create formatter
        formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )
        console_handler.setFormatter(formatter)

        # Add console handler to logger
        logger.addHandler(console_handler)

    return logger



from typing import Dict
from phenex.mappers import DomainsDictionary
import pandas as pd
import numpy as np
from dataclasses import asdict


def generate_mock_mapped_tables(
    n_patients: int, domains: DomainsDictionary
) -> Dict[str, pd.DataFrame]:
    """
    Generate fake data for N patients based on the given domains.

    Args:
        n_patients (int): The number of patients to generate data for.
        domains (DomainsDictionary): The domains dictionary containing the table mappers.

    Returns:
        Dict[str, pd.DataFrame]: A dictionary where keys are domain names and values are DataFrames with fake data.
    """
    fake_data = {}
    for domain, mapper in domains.domains_dict.items():
        columns = [field for field in asdict(mapper).keys() if field != "NAME_TABLE"]
        data = {}
        for col in columns:
            if "DATE" in col:
                start_date = pd.to_datetime("2000-01-01")
                end_date = pd.to_datetime("2020-12-31")
                data[col] = pd.to_datetime(
                    np.random.randint(start_date.value, end_date.value, n_patients)
                ).date
            elif "ID" in col:
                data[col] = np.arange(1, n_patients + 1)
            elif "VALUE" in col:
                data[col] = np.random.uniform(0, 100, n_patients)
            elif "CODE_TYPE" in col:
                if "CONDITION" in domain:
                    data[col] = np.random.choice(["ICD-10", "SNOMED"], n_patients)
                elif "DRUG" in domain:
                    data[col] = np.random.choice(["NDC", "RxNorm"], n_patients)
                elif "PROCEDURE" in domain:
                    data[col] = np.random.choice(["CPT", "HCPCS"], n_patients)
                else:
                    data[col] = np.random.choice(["TYPE1", "TYPE2"], n_patients)
            elif "CODE" in col:
                data[col] = np.random.choice(
                    ["A", "B", "C", "D", "E", "F", "G"], n_patients
                )
            else:
                data[col] = np.random.choice(range(1000), n_patients)
        fake_data[domain] = pd.DataFrame(data)
    return fake_data



from phenex.filters.filter import Filter
from typing import List, Optional, Union, Dict


class CategoricalFilter(Filter):
    """
    This class filters events in an EventTable based on specified categorical values.

    Attributes:
        column_name (str): The name of the column to filter by.
        allowed_values (List[Union[str, int]]): The list of allowed values for the column.
        domain (Optional[str]): The domain to which the filter applies.
        operator (Optional[str]): The operator used for filtering. Options are "isin", "notin", "isnull", "notnull". Default is "isin". Be aware that "notin" removes null values! Thus "notin" can better be described 'has a value and and is not in'

    Methods:
        filter(table: PhenexTable) -> PhenexTable:
            Filters the given PhenexTable based on the specified column and allowed values.
            Parameters:
                table (PhenexTable): The table containing events to be filtered.
            Returns:
                PhenexTable: The filtered PhenexTable with events matching the allowed values.

        autojoin_filter(table: PhenexTable, tables: dict = None) -> PhenexTable:
            Automatically joins the necessary tables and applies the filter. Use when the input table does not contain the column that defines the filter. For this to work, the tables must specify all required join keys. See DomainsDictionary for details.
            Parameters:
                table (PhenexTable): The table containing events to be filtered.
                tables (dict): A dictionary of tables for joining.
            Returns:
                PhenexTable: The filtered PhenexTable with events matching the allowed values.

    Examples:
        ```
        # Example 1: Filter for SEX = 'Female'
        sex_filter = CategoricalFilter(
            column_name="SEX",
            allowed_values=["Female"],
            domain="PERSON"
        )
        ```

        ```
        # Example 2: Filter for inpatient (domain = encounter)
        inpatient_filter = CategoricalFilter(
            column_name="ENCOUNTER_TYPE",
            allowed_values=["INPATIENT"],
            domain="ENCOUNTER"
        )
        ```

        ```
        # Example 3: Filter for primary diagnosis position
        primary_diagnosis_filter = CategoricalFilter(
            column_name="DIAGNOSIS_POSITION",
            allowed_values=[1],
            domain="DIAGNOSIS"
        )
        ```

        ```
        # Example 4: Applying multiple filters in combination
        inpatient_primary_position = inpatient_filter & primary_diagnosis_filter
        ```

        ```
        # Example 5: Filter for not null status
        primary_diagnosis_filter = CategoricalFilter(
            column_name="DIAGNOSIS_STATUS",
            operator="notnull",
            domain="DIAGNOSIS"
        )
        ```

        ```
        # Example 6: Filter for diagnosis status. Notice that this will remove null rows! Only patients with a value in the diagnosis_status that is NOT history_of will remain.
        primary_diagnosis_filter = CategoricalFilter(
            column_name="DIAGNOSIS_STATUS",
            operator="notin",
            allowed_values=["HISTORY_OF"],
            domain="DIAGNOSIS"
        )
        ```
    """

    def __init__(
        self,
        column_name: str = None,
        allowed_values: Optional[List[Union[str, int]]] = None,
        domain: Optional[str] = None,
        operator: Optional[str] = "isin",
    ):
        self.column_name = column_name
        self.allowed_values = allowed_values
        self.domain = domain
        self.operator = operator
        if operator not in ["isin", "notin", "isnull", "notnull"]:
            raise ValueError(
                f"Invalid operator '{operator}'. Must be one of: 'isin', 'notin', 'isnull', 'notnull'."
            )
        if operator in ["isnull", "notnull"] and allowed_values is not None:
            raise ValueError(
                f"Allowed values should be None for operator '{operator}'."
            )
        super(CategoricalFilter, self).__init__()

    def _filter(self, table: "PhenexTable"):
        # if simply setting the desired column with the categorical filter, return the table without any filtering. This occurs when allowed_values is null, or the operator is 'isnull' or 'notnull'
        if self.allowed_values is None and "null" not in self.operator:
            return table
        if self.operator == "isin" and self.allowed_values:
            table = table.filter(table[self.column_name].isin(self.allowed_values))
        if self.operator == "notin" and self.allowed_values:
            table = table.filter(~table[self.column_name].isin(self.allowed_values))
        if self.operator == "isnull":
            table = table.filter(table[self.column_name].isnull())
        if self.operator == "notnull":
            table = table.filter(~table[self.column_name].isnull())
        return table

    def autojoin_filter(
        self, table: "PhenexTable", tables: Optional[Dict[str, "PhenexTable"]] = None
    ) -> "PhenexTable":
        if self.column_name not in table.columns:
            if self.domain not in tables.keys():
                raise ValueError(
                    f"Table required for categorical filter ({self.domain}) does not exist within domains dicitonary"
                )
            table = table.join(tables[self.domain], domains=tables)
            # TODO downselect to original columns
        return self._filter(table)



from ibis.expr.types.relations import Table

from typing import Union
import ibis
import copy


class PhenexTable:
    """
    Phenex provides certain table types on which it knows how to operate. For instance, Phenex implements a CodeTable, which is an event table containing codes. Phenex has abstracted operations for each table type. For instance, given a CodeTable, Phenex knows how to filter this table based on the presence of codes within that table. Phenex doesn't care if the code table is actually a diagnosis code table or a procedure code table or a medication code table.

    In onboarding a new data model to Phenex, the tables must be mapped into Phenex table types by subclassing the appropriate PhenexTable. When subclassing a PhenexTable, you must define:

        1. COLUMN_MAPPING: a mapping of the input table columns to the fields on the chosen PhenexTable type (e.g. 'CD' maps to 'CODE' in a CodeTable).
        2. JOIN_KEYS: if you want to use the autojoin functionality of PhenexTable, you must specify what keys to use to join pairs of tables
        3. PATHS: if you want to use the autojoin functionality of PhenexTable for more complex joins, you must specify join paths to take to get from one table to another

    Note that for each table type, there are REQUIRED_FIELDS, i.e., fields that MUST be defined for Phenex to work with such a table and KNOWN_FIELDS, i.e., fields that Phenex internally understands what to do with (there is a Phenotype that knows how to work with that field). For instance, in a PhenexPersonTable, one MUST define PERSON_ID, but DATE_OF_BIRTH is an optional field that PhenEx can process if given and transform into AGE. These are fixed for each table type and should not be overridden.
    """

    NAME_TABLE = "PHENEX_TABLE"  # name of table in the database
    JOIN_KEYS = {}  # dict: class name -> List[phenex column names]
    KNOWN_FIELDS = []  # List[phenex column names]
    DEFAULT_MAPPING = {}  # dict: input column name -> phenex column name
    PATHS = {}  # dict: table class name -> List[other table class names]
    REQUIRED_FIELDS = list(DEFAULT_MAPPING.keys())

    def __init__(self, table, name=None, column_mapping={}):
        """
        Instantiate a PhenexTable, possibly overriding NAME_TABLE and COLUMN_MAPPING.
        """

        if not isinstance(table, Table):
            raise TypeError(
                f"Cannot instantiatiate {self.__class__.__name__} from {type(table)}. Must be ibis Table."
            )

        self.NAME_TABLE = name or self.NAME_TABLE

        self.column_mapping = self._get_column_mapping(column_mapping)
        self._table = table.mutate(**self.column_mapping)

        for key in self.REQUIRED_FIELDS:
            try:
                getattr(self._table, key)
            except AttributeError:
                raise ValueError(f"Required field {key} not defined in COLUMN_MAPPING.")

        self._add_phenotype_table_relationship()

    def _add_phenotype_table_relationship(self):
        self.JOIN_KEYS["PhenotypeTable"] = ["PERSON_ID"]

    def _get_column_mapping(self, column_mapping=None):
        column_mapping = column_mapping or {}
        for key in column_mapping.keys():
            if key not in self.KNOWN_FIELDS:
                raise ValueError(
                    f"Unknown mapped field {key} --> {column_mapping[key]} for f{type(self)}."
                )
        default_mapping = copy.deepcopy(self.DEFAULT_MAPPING)
        default_mapping.update(column_mapping)
        return default_mapping

    def __getattr__(self, name):
        # pass all attributes on to underlying table
        return getattr(self._table, name)

    def __getitem__(self, key):
        return self._table[key]

    @property
    def table(self):
        return self._table

    def join(self, other: "PhenexTable", *args, domains=None, **kwargs):
        """
        The join method performs a join of PhenexTables, using autojoin functionality if Phenex is able to find the table types specified in PATHS.
        """
        if isinstance(other, Table):
            return type(self)(self.table.join(other, *args, **kwargs))

        if not isinstance(other, PhenexTable):
            raise TypeError(f"Expected a PhenexTable instance, got {type(other)}")
        if len(args):
            # if user specifies join keys and join type, simply perform join as specified
            return type(self)(self.table.join(other.table, *args, **kwargs))

        # Do an autojoin by finding a path from the left to the right table and sequentially joining as necessary
        # joined table is the sequentially joined table
        # current table is the table for the left join in the current iteration
        joined_table = current_left_table = self
        for right_table_class_name in self._find_path(other):
            # get the next right table
            right_table_search_results = [
                v
                for k, v in domains.items()
                if v.__class__.__name__ == right_table_class_name
            ]
            if len(right_table_search_results) != 1:
                raise ValueError(
                    f"Unable to find unqiue {right_table_class_name} required to join {other.__class__.__name__}"
                )
            right_table = right_table_search_results[0]
            print(
                f"\tJoining : {current_left_table.__class__.__name__} to {right_table.__class__.__name__}"
            )

            # join keys are defined by the left table; in theory should enforce symmetry
            join_keys = current_left_table.JOIN_KEYS[right_table_class_name]
            columns = list(set(joined_table.columns + right_table.columns))
            # subset columns, making sure to set type of table to the very left table (self)
            joined_table = type(self)(
                joined_table.join(right_table, join_keys, **kwargs).select(columns)
            )
            current_left_table = right_table
        return joined_table

    def mutate(self, *args, **kwargs):
        return type(self)(self.table.mutate(*args, **kwargs), name=self.NAME_TABLE)

    def _find_path(self, other):
        start_name = self.__class__.__name__
        end_name = other.__class__.__name__
        # first see if direct connection
        try:
            self.JOIN_KEYS[end_name]
            return [end_name]
        except KeyError:
            try:
                return self.PATHS[end_name] + [end_name]
            except KeyError:
                raise ValueError(
                    f"Cannot autojoin {start_name} --> {end_name}. Please specify join path in PATHS."
                )

    def filter(self, expr):
        """
        Filter the table by an Ibis Expression or using a PhenExFilter.
        """
        input_columns = self.columns
        if isinstance(expr, ibis.expr.types.Expr) or isinstance(expr, list):
            filtered_table = self.table.filter(expr)
        else:
            filtered_table = expr.filter(self)

        return type(self)(
            filtered_table.select(input_columns),
            name=self.NAME_TABLE,
            column_mapping=self.column_mapping,
        )


class PhenexPersonTable(PhenexTable):
    NAME_TABLE = "PERSON"
    JOIN_KEYS = {
        "CodeTable": ["PERSON_ID"],
        "PhenexVisitOccurrenceTable": ["PERSON_ID"],
    }
    KNOWN_FIELDS = [
        "PERSON_ID",
        "DATE_OF_BIRTH",
        "YEAR_OF_BIRTH",
        "DATE_OF_DEATH",
        "SEX",
        "ETHNICITY",
    ]
    DEFAULT_MAPPING = {"PERSON_ID": "PERSON_ID"}


class EventTable(PhenexTable):
    NAME_TABLE = "EVENT"
    KNOWN_FIELDS = ["PERSON_ID", "EVENT_DATE"]
    DEFAULT_MAPPING = {
        "PERSON_ID": "PERSON_ID",
        "EVENT_DATE": "EVENT_DATE",
    }


class CodeTable(PhenexTable):
    NAME_TABLE = "CODE"
    RELATIONSHIPS = {
        "PhenexPersonTable": ["PERSON_ID"],
        "PhenexVisitOccurrenceTable": ["PERSON_ID", "VISIT_DETAIL_ID"],
    }
    KNOWN_FIELDS = ["PERSON_ID", "EVENT_DATE", "CODE", "CODE_TYPE", "VISIT_DETAIL_ID"]
    DEFAULT_MAPPING = {
        "PERSON_ID": "PERSON_ID",
        "EVENT_DATE": "EVENT_DATE",
        "CODE": "CODE",
    }


class PhenexVisitOccurrenceTable(PhenexTable):
    NAME_TABLE = "VISIT_OCCURRENCE"
    RELATIONSHIPS = {
        "PhenexPersonTable": ["PERSON_ID"],
        "CodeTable": ["PERSON_ID", "VISIT_OCCURRENCE_ID"],
    }
    KNOWN_FIELDS = [
        "PERSON_ID",
        "VISIT_OCCURRENCE_ID",
        "VISIT_OCCURRENCE_SOURCE_VALUE",
    ]

    DEFAULT_MAPPING = {
        "PERSON_ID": "PERSON_ID",
        "VISIT_OCCURRENCE_ID": "VISIT_DETAIL_ID",
        "VISIT_OCCURRENCE_SOURCE_VALUE": "VISIT_DETAIL_SOURCE_VALUE",
    }


class PhenexIndexTable(PhenexTable):
    NAME_TABLE = "INDEX"
    JOIN_KEYS = {
        "CodeTable": ["PERSON_ID"],
        "PhenexVisitOccurrenceTable": ["PERSON_ID"],
    }
    KNOWN_FIELDS = [
        "PERSON_ID",
        "INDEX_DATE",
    ]
    DEFAULT_MAPPING = {"PERSON_ID": "PERSON_ID", "INDEX_DATE": "INDEX_DATE"}


class PhenexObservationPeriodTable(PhenexTable):
    NAME_TABLE = "OBSERVATION_PERIOD"
    KNOWN_FIELDS = [
        "PERSON_ID",
        "START_DATE",
        "END_DATE",
    ]
    DEFAULT_MAPPING = {
        "PERSON_ID": "PERSON_ID",
        "START_DATE": "OBSERVATION_PERIOD_START_DATE",
        "END_DATE": "OBSERVATION_PERIOD_END_DATE",
    }


class MeasurementTable(PhenexTable):
    NAME_TABLE = "MEASUREMENT"
    RELATIONSHIPS = {
        "PhenexPersonTable": ["PERSON_ID"],
        "PhenexVisitOccurrenceTable": ["PERSON_ID", "VISIT_DETAIL_ID"],
    }
    KNOWN_FIELDS = [
        "PERSON_ID",
        "EVENT_DATE",
        "CODE",
        "CODE_TYPE",
        "VISIT_DETAIL_ID",
        "VALUE",
    ]
    DEFAULT_MAPPING = {
        "PERSON_ID": "PERSON_ID",
        "EVENT_DATE": "EVENT_DATE",
        "CODE": "CODE",
        "VALUE": "VALUE",
    }


class PhenotypeTable(PhenexTable):
    NAME_TABLE = "PHENOTYPE"
    KNOWN_FIELDS = ["PERSON_ID", "BOOLEAN", "EVENT_DATE", "VALUE"]
    DEFAULT_MAPPING = {
        "PERSON_ID": "PERSON_ID",
        "BOOLEAN": "BOOLEAN",
        "EVENT_DATE": "EVENT_DATE",
        "VALUE": "VALUE",
    }


def is_phenex_person_table(table: PhenexTable) -> bool:
    """
    Check if given table is a person table.
    One could check one row per patient?
    """
    return set(table.columns) >= set(PhenexPersonTable.REQUIRED_FIELDS)


def is_phenex_code_table(table: PhenexTable) -> bool:
    """
    Check if given table is a code table.
    """
    return set(table.columns) >= set(CodeTable.REQUIRED_FIELDS)


def is_phenex_event_table(table: PhenexTable) -> bool:
    """
    Check if given table is a code table.
    """
    return set(table.columns) >= set(EventTable.REQUIRED_FIELDS)


def is_phenex_phenotype_table(table: PhenexTable) -> bool:
    """
    Check if given table is a code table.
    """
    return set(table.columns) >= set(PhenotypeTable.REQUIRED_FIELDS)


def is_phenex_index_table(table: PhenexTable) -> bool:
    """
    Check if given table is a code table.
    """
    return set(table.columns) >= set(PhenexIndexTable.REQUIRED_FIELDS)


PHENOTYPE_TABLE_COLUMNS = ["PERSON_ID", "BOOLEAN", "EVENT_DATE", "VALUE"]



import ibis
from ibis import _

from phenex.phenotypes.phenotype import Phenotype
from phenex.filters.relative_time_range_filter import RelativeTimeRangeFilter
from phenex.filters import DateFilter, ValueFilter
from phenex.tables import is_phenex_code_table, PHENOTYPE_TABLE_COLUMNS, PhenotypeTable
from phenex.aggregators import First, Last

from phenex.util import create_logger

logger = create_logger(__name__)


class BinPhenotype(Phenotype):
    """
    BinPhenotype converts numeric values into categorical bin labels. To use, pass it a numeric valued phenotype such as AgePhenotype, MeasurementPhenotype, ArithmeticPhenotype, or ScorePhenotype.

    Takes a phenotype that returns numeric values (like age, measurements, etc.)
    and converts the VALUE column into bin labels like "[10-20)", "[20-30)", etc.

    DATE: The event date selected from the input phenotype
    VALUE: A categorical variable representing the bin label that the numeric value falls into

    Parameters:
        name: The name of the phenotype.
        phenotype: The phenotype that returns numeric values of interest (AgePhenotype, MeasurementPhenotype, etc.)
        bins: List of bin edges. Default is [0, 10, 20, ..., 100] for age ranges.

    Example:
        ```python
        # Create an age phenotype
        age = AgePhenotype()

        # Create bins for age groups: [0-10), [10-20), [20-30), etc.
        binned_age = BinPhenotype(
            name="age_groups",
            phenotype=age,
            bins=[0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100]
        )

        tables = {"PERSON": example_person_table}
        result_table = binned_age.execute(tables)

        # Result will have VALUE column with labels like "[20-30)", "[30-40)", etc.
        display(result_table)
        ```
    """

    def __init__(
        self,
        phenotype: Phenotype,
        bins=list(range(0, 101, 10)),
        **kwargs,
    ):
        super(BinPhenotype, self).__init__(**kwargs)
        self.bins = bins
        self.phenotype = phenotype
        if self.phenotype.__class__.__name__ not in [
            "AgePhenotype",
            "MeasurementPhenotype",
            "ArithmeticPhenotype",
            "ScorePhenotype",
        ]:
            raise ValueError(
                f"Invalid phenotype type: {self.phenotype.__class__.__name__}"
            )
        self.children = [phenotype]

    def _execute(self, tables) -> PhenotypeTable:
        # Execute the child phenotype to get the initial table to filter
        table = self.phenotype.table

        # Create bin labels
        bin_labels = []

        # Add a bin for values < first bin edge
        bin_labels.append(f"<{self.bins[0]}")

        # Add bins for each range
        for i in range(len(self.bins) - 1):
            bin_labels.append(f"[{self.bins[i]}-{self.bins[i+1]})")

        # Add a final bin for values >= last bin edge
        bin_labels.append(f">={self.bins[-1]}")

        # Create binning logic using Ibis case statements
        value_col = table.VALUE

        # Start with the case expression
        case_expr = None

        # Handle values < first bin edge
        first_condition = value_col < self.bins[0]
        case_expr = ibis.case().when(first_condition, bin_labels[0])

        # Create conditions for each bin range
        for i in range(len(self.bins) - 1):
            condition = (value_col >= self.bins[i]) & (value_col < self.bins[i + 1])
            case_expr = case_expr.when(condition, bin_labels[i + 1])

        # Handle values >= last bin edge
        final_condition = value_col >= self.bins[-1]
        case_expr = case_expr.when(final_condition, bin_labels[-1])

        # Handle null values
        case_expr = case_expr.else_(None)

        # Replace the VALUE column with bin labels
        table = table.mutate(VALUE=case_expr.end())

        return table



from typing import Dict, Optional
from ibis.expr.types.relations import Table
import ibis
import pandas as pd

from phenex.derived_tables.derived_table import DerivedTable
from phenex.util import create_logger
from phenex.tables import PhenexTable

logger = create_logger(__name__)


class CombineOverlappingPeriods(DerivedTable):
    """
    CombineOverlappingPeriods takes overlapping and consecutive time periods the source table and combines them into a single time period with a single start and end date on a per patient level. For example, if a patient has two visits with the same start and end date, they will be combined into one visit. If a patient has two visits with overlapping dates, they will be combined into one visit with the earliest start date and the latest end date. If a patient has two visits with consecutive dates, they will be combined into one visit with the earliest start date and the latest end date.
    This is useful for creating a single time period for a patient, e.g. admission discharge periods, vaccination periods, etc. It is also useful for creating a single time period for a patient when there are multiple visits with the same start and end date, or overlapping dates.
    """

    def __init__(
        self, categorical_filter: Optional["CategoricalFilter"] = None, **kwargs
    ):
        self.categorical_filter = categorical_filter
        super(CombineOverlappingPeriods, self).__init__(**kwargs)

    def execute(
        self,
        tables: Dict[str, Table],
    ) -> "PhenexTable":
        # get the appropriate table
        table = tables[self.source_domain]
        logger.warning(
            "CombineOverlappingTables has known potential performance issues especially when working with large cohorts. Please open an issue on GitHub if performance issues appear."
        )
        if self.categorical_filter is not None:
            # apply the categorical filter to the table
            table = self.categorical_filter.autojoin_filter(table, tables)

        # subset to only the relevant columns
        df = table.select("PERSON_ID", "START_DATE", "END_DATE").to_pandas()

        # perform time period merging
        df = df.sort_values(["PERSON_ID", "START_DATE", "END_DATE"])
        result = []
        for pid, group in df.groupby("PERSON_ID"):
            intervals = group[["START_DATE", "END_DATE"]].values
            merged = []
            for start, end in intervals:
                if not merged or start > merged[-1][1] + pd.Timedelta(days=1):
                    merged.append([start, end])
                else:
                    merged[-1][1] = max(merged[-1][1], end)
            for start, end in merged:
                result.append({"PERSON_ID": pid, "START_DATE": start, "END_DATE": end})
        df_result = pd.DataFrame(result)
        # create a new table with the merged results
        table = ibis.memtable(df_result)
        return PhenexTable(table, name=self.dest_domain)



from typing import Dict, Optional

import ibis
from ibis.expr.types.relations import Table

from phenex.phenotypes.phenotype import Phenotype
from phenex.filters import ValueFilter, Value
from phenex.tables import PhenotypeTable, is_phenex_person_table
from phenex.filters.relative_time_range_filter import RelativeTimeRangeFilter
from phenex.util import create_logger

logger = create_logger(__name__)


class AgePhenotype(Phenotype):
    """
    AgePhenotype is a class that represents an age-based phenotype. It calculates the age of individuals
    based on their date of birth and an optional anchor phenotype. The age is computed in years and can
    be filtered within a specified range.

    Parameters:
        name: Name of the phenotype, default is 'age'.
        value_filter: Filter the returned patients based on their age (in years)
        anchor_phenotype: An optional anchor phenotype to calculate relative age.
        domain: Domain of the phenotype, default is 'PERSON'.

    Attributes:
        table (PhenotypeTable): The resulting phenotype table after filtering (None until execute is called)

    Example: Age at First Atrial Fibrillation Diagnosis
        ```python
        from phenex.phenotypes import CodelistPhenotype
        from phenex.codelists import Codelist

        af_codelist = Codelist([313217])
        af_phenotype = CodelistPhenotype(
            name="af",
            domain='CONDITION_OCCURRENCE',
            codelist=af_codelist,
            return_date='first',
        )

        age_phenotype = AgePhenotype(
            value_filter=ValueFilter(
                min_value=GreaterThan(18),
                max_value=LessThan(65)
                ),
            anchor_phenotype=af_phenotype
        )

        result_table = age_phenotype.execute(tables)
        display(result_table)
        ```
    """

    # FIXME this will become a problem when modern medicine allows people to live more
    # than 365*4 years (so they accumulate enough leap days to get an extra year)
    DAYS_IN_YEAR = 365

    def __init__(
        self,
        name: Optional[str] = "AGE",
        value_filter: Optional[ValueFilter] = None,
        anchor_phenotype: Optional[Phenotype] = None,
        domain: str = "PERSON",
        **kwargs,
    ):
        self.min_age = self.max_age = None
        if value_filter:
            self.min_age = value_filter.min_value
            self.max_age = value_filter.max_value

        self.domain = domain
        self.anchor_phenotype = anchor_phenotype

        self.time_range_filter = RelativeTimeRangeFilter(
            anchor_phenotype=anchor_phenotype
        )

        # Set children to the dependent PHENOTYPES
        if anchor_phenotype is not None:
            self.children = [anchor_phenotype]
        else:
            self.children = []

        super(AgePhenotype, self).__init__(name=name, **kwargs)

    def _execute(self, tables: Dict[str, Table]) -> PhenotypeTable:
        person_table = tables[self.domain]
        assert is_phenex_person_table(person_table)

        if "YEAR_OF_BIRTH" in person_table.columns:
            if "DATE_OF_BIRTH" in person_table.columns:
                logger.debug(
                    "Year of birth and date of birth is present, taking date of birth where possible otherwise setting date of birth to june 6th"
                )
                date_of_birth = ibis.coalesce(
                    ibis.date(person_table.DATE_OF_BIRTH),
                    ibis.date(person_table.YEAR_OF_BIRTH, 6, 1),
                )
            else:
                logger.debug(
                    "Only year of birth is present in person table, setting birth date to june 6th"
                )
                date_of_birth = ibis.date(person_table.YEAR_OF_BIRTH, 6, 1)
        else:
            logger.debug("Year of birth not present, taking date of birth")
            date_of_birth = ibis.date(person_table.DATE_OF_BIRTH)
        person_table = person_table.mutate(EVENT_DATE=date_of_birth)

        # Apply the time range filter
        table = person_table
        if self.anchor_phenotype is not None:
            if self.anchor_phenotype.table is None:
                raise ValueError(
                    f"Dependent Phenotype {self.anchor_phenotype.name} must be executed before this node can run!"
                )
            else:
                anchor_table = self.anchor_phenotype.table
                reference_column = anchor_table.EVENT_DATE
                # Note that joins can change column names if the tables have name collisions!
                table = table.join(anchor_table, "PERSON_ID")
        else:
            assert (
                "INDEX_DATE" in table.columns
            ), f"INDEX_DATE column not found in table {table}"
            reference_column = table.INDEX_DATE

        YEARS_FROM_ANCHOR = (
            reference_column.delta(table.EVENT_DATE, "day") / self.DAYS_IN_YEAR
        ).floor()
        table = table.mutate(YEARS_FROM_ANCHOR=YEARS_FROM_ANCHOR)

        conditions = []
        # Fix this, this logic needs to be abstracted to a ValueFilter
        if self.min_age is not None:
            if self.min_age.operator == ">":
                conditions.append(table.YEARS_FROM_ANCHOR > self.min_age.value)
            elif self.min_age.operator == ">=":
                conditions.append(table.YEARS_FROM_ANCHOR >= self.min_age.value)
            else:
                raise ValueError("Operator for min days be > or >=")
        if self.max_age is not None:
            if self.max_age.operator == "<":
                conditions.append(table.YEARS_FROM_ANCHOR < self.max_age.value)
            elif self.max_age.operator == "<=":
                conditions.append(table.YEARS_FROM_ANCHOR <= self.max_age.value)
            else:
                raise ValueError("Operator for max days be < or <=")
        if conditions:
            table = table.filter(conditions)
        person_table = table

        person_table = person_table.mutate(VALUE=person_table.YEARS_FROM_ANCHOR)

        return person_table



from typing import Dict, Union, Optional
from datetime import date
from ibis.expr.types.relations import Table
import ibis
from phenex.tables import PhenotypeTable, PHENOTYPE_TABLE_COLUMNS
from phenex.phenotypes.phenotype import Phenotype, ComputationGraph
from phenex.phenotypes.functions import hstack


class ComputationGraphPhenotype(Phenotype):
    """
    ComputationGraphPhenotypes are a type of CompositePhenotype that performs computations using phenotypes. The ComputationGraphPhenotype is a base class and must be subclassed.
    The subclasses of ComputationGraphPhenotype differ depending on which columns of the component phenotype tables are used as input and where the output is placed in the output phenotype table. The two options for both input and output are 'boolean' or 'value'.

    ## Comparison table of CompositePhenotype classes
    +---------------------+-------------+------------+
    |                     | Operates on | Populates  |
    +=====================+=============+============+
    | ArithmeticPhenotype | value       | value      |
    +---------------------+-------------+------------+
    | LogicPhenotype      | boolean     | boolean    |
    +---------------------+-------------+------------+
    | ScorePhenotype      | boolean     | value      |
    +---------------------+-------------+------------+

    Parameters:
        expression: The arithmetic expression to be evaluated composed of phenotypes combined by python arithmetic operations.
        return_date: The date to be returned for the phenotype. Can be "first", "last", or a Phenotype object.
        operate_on: The column to operate on. Can be "boolean" or "value".
        populate: The column to populate. Can be "boolean" or "value".
        reduce: Whether to reduce the phenotype table to only include rows where the boolean column is True. This is only relevant if populate is "boolean".

    Attributes:
        table (PhenotypeTable): The resulting phenotype table after filtering (None until execute is called)
    """

    def __init__(
        self,
        expression: ComputationGraph,
        return_date: Union[str, Phenotype],
        name: Optional[str] = None,
        aggregation_index=["PERSON_ID"],
        operate_on: str = "boolean",
        populate: str = "value",
        reduce: bool = False,
        **kwargs,
    ):
        if name is None:
            name = str(expression)
        super(ComputationGraphPhenotype, self).__init__(name=name, **kwargs)
        self.expression = expression
        self.return_date = return_date
        self.aggregation_index = aggregation_index
        self.operate_on = operate_on
        self.populate = populate
        self.reduce = reduce
        self.children = self.expression.get_leaf_phenotypes()

    def _execute(self, tables: Dict[str, Table]) -> PhenotypeTable:
        """
        Executes the score phenotype processing logic.

        Args:
            tables (Dict[str, Table]): A dictionary where the keys are table names and the values are Table objects.

        Returns:
            PhenotypeTable: The resulting phenotype table containing the required columns.
        """
        joined_table = hstack(self.children, tables["PERSON"].select("PERSON_ID"))

        if self.populate == "value" and self.operate_on == "boolean":
            for child in self.children:
                column_name = f"{child.name}_BOOLEAN"
                mutated_column = ibis.ifelse(
                    joined_table[column_name].isnull(),
                    0,
                    joined_table[column_name].cast("int"),
                ).cast("float")

                joined_table = joined_table.mutate(**{column_name: mutated_column})

        if self.populate == "value":
            _expression = self.expression.get_value_expression(
                joined_table, operate_on=self.operate_on
            )
            joined_table = joined_table.mutate(VALUE=_expression)
            # Arithmetic operations imply a boolean 'and' of children i.e. child1 + child two implies child1 and child2. if there are any null values in value calculations this is because one of the children is null, so we filter them out as the implied boolean condition is not met.
            joined_table = joined_table.filter(joined_table["VALUE"].notnull())

        elif self.populate == "boolean":
            _expression = self.expression.get_boolean_expression(
                joined_table, operate_on=self.operate_on
            )
            joined_table = joined_table.mutate(BOOLEAN=_expression)

        # Return the first or last event date
        date_columns = self._coalesce_all_date_columns(joined_table)
        if self.return_date == "first":
            joined_table = joined_table.mutate(EVENT_DATE=ibis.least(*date_columns))
        elif self.return_date == "last":
            joined_table = joined_table.mutate(EVENT_DATE=ibis.greatest(*date_columns))
        elif self.return_date == "all":
            joined_table = self._return_all_dates(joined_table, date_columns)
        elif isinstance(self.return_date, Phenotype):
            joined_table = joined_table.mutate(
                EVENT_DATE=getattr(joined_table, f"{self.return_date.name}_EVENT_DATE")
            )
        else:
            joined_table = joined_table.mutate(EVENT_DATE=ibis.null(date))

        # Reduce the table to only include rows where the boolean column is True
        if self.reduce:
            joined_table = joined_table.filter(joined_table.BOOLEAN == True)

        # Add a null value column if it doesn't exist, for example in the case of a LogicPhenotype
        schema = joined_table.schema()
        if "VALUE" not in schema.names:
            joined_table = joined_table.mutate(VALUE=ibis.null())

        return joined_table

    def _return_all_dates(self, table, date_columns):
        """
        If return date = all, we want to return all the dates on which phenotype criteria are fulfilled; this is a union of all the non-null dates in any leaf phenotype date columns.

        Args:
            table: The Ibis table object (e.g., joined_table) that contains all leaf phenotypes stacked horizontally
            date_columns: List of base columns as ibis objects

        Returns:
            Ibis expression representing the UNION of all non null dates.
        """
        # get all the non-null dates for each date column
        non_null_dates_by_date_col = []
        for date_col in date_columns:
            non_null_dates = table.filter(date_col.notnull()).mutate(
                EVENT_DATE=date_col
            )
            non_null_dates_by_date_col.append(non_null_dates)

        # do the union of all the non-null dates
        all_dates = non_null_dates_by_date_col[0]
        for non_null_dates in non_null_dates_by_date_col[1:]:
            all_dates = all_dates.union(non_null_dates)
        return all_dates

    def _coalesce_all_date_columns(self, table):
        """
        ComputationGraphPhenotypes have multiple possible date columns. To work with these date columns, which may be null, we perform a coalesce operation for each date column, which allows operations such as 'least' and 'greatest' to work correctly.

        Args:
            table: The Ibis table object (e.g., joined_table).

        Returns:
            Ibis expression representing the COALESCE of the columns.
        """
        coalesce_expressions = []

        names = [col for col in table.columns if "EVENT_DATE" in col]

        # computation graph phenotypes occasionally have a single phenotype in the expression e.g. negation of a single phenotype. In this case, some backends (e.g. Snowflake) do not work with coalesce of a single column. Therefore, if only a single phenotype, return the select of that column
        if len(names) == 1:
            return [table[names[0]]]

        # if more than one phenotype in expression, coalesce return date columns to allow for further date selection
        for i in range(len(names)):
            rotated_names = names[i:] + names[:i]
            coalesce_expr = ibis.coalesce(
                *(getattr(table, col) for col in rotated_names)
            )
            coalesce_expressions.append(coalesce_expr)

        return coalesce_expressions


class ScorePhenotype(ComputationGraphPhenotype):
    """
    ScorePhenotype is a CompositePhenotype that performs arithmetic operations using the **boolean** column of its component phenotypes and populations the **value** column. It should be used for calculating medical scores such as CHADSVASC, HASBLED, etc.

    --> See the comparison table of CompositePhenotype classes

    Parameters:
        expression: The arithmetic expression to be evaluated composed of phenotypes combined by python arithmetic operations.
        return_date: The date to be returned for the phenotype. Can be "first", "last", or a Phenotype object.
        name: The name of the phenotype.

    Attributes:
        table (PhenotypeTable): The resulting phenotype table after filtering (None until execute is called)

    Example:
    ```python
    # Create component phenotypes individually
    hypertension = Phenotype(Codelist('hypertension'))
    hf = Phenotype(Codelist('chf'))
    age_gt_45 = AgePhenotype(min_age=GreaterThan(45))

    # Create the ScorePhenotype that defines a score which is 2*age + 1 if
    # hypertension or chf are present, respectively. Notice that the boolean
    # column of the component phenotypes are used for calculation and the value
    # column is populated of the ScorePhenotype table.
    pt = ScorePhenotype(
        expression = 2 * age_gt_45 + hypertension + chf,
    )
    ```
    """

    def __init__(
        self,
        expression: ComputationGraph,
        return_date: Union[str, Phenotype] = "first",
        name: str = None,
        **kwargs,
    ):
        super(ScorePhenotype, self).__init__(
            name=name,
            expression=expression,
            return_date=return_date,
            operate_on="boolean",
            populate="value",
        )


class ArithmeticPhenotype(ComputationGraphPhenotype):
    """
    ArithmeticPhenotype is a composite phenotype that performs arithmetic operations using the **value** column of its component phenotypes and populations the **value** column. It should be used for calculating values such as BMI, GFR or converting units.
    --> See the comparison table of CompositePhenotype classes

    Parameters:
        expression: The arithmetic expression to be evaluated composed of phenotypes combined by python arithmetic operations.
        return_date: The date to be returned for the phenotype. Can be "first", "last", or a Phenotype object.
        name: The name of the phenotype.

    Attributes:
        table (PhenotypeTable): The resulting phenotype table after filtering (None until execute is called)

    Example:
    ```python
    # Create component phenotypes individually
    height = MeasurementPhenotype(Codelist('height'))
    weight = MeasurementPhenotype(Codelist('weight'))

    # Create the ArithmeticPhenotype that defines the BMI score
    bmi = ArithmeticPhenotype(
        expression = weight / height**2,
    )
    ```
    """

    def __init__(
        self,
        expression: ComputationGraph,
        return_date: Union[str, Phenotype] = "first",
        name: str = None,
        **kwargs,
    ):
        super(ArithmeticPhenotype, self).__init__(
            name=name,
            expression=expression,
            return_date=return_date,
            operate_on="value",
            populate="value",
        )


class LogicPhenotype(ComputationGraphPhenotype):
    """
    LogicPhenotype is a composite phenotype that performs boolean operations using the **boolean** column of its component phenotypes and populations the **boolean** column of the resulting phenotype table. It should be used in any instance where multiple phenotypes are logically combined, for example, does a patient have diabetes AND hypertension, etc.

    --> See the comparison table of CompositePhenotype classes

    Parameters:
        expression: The logical expression to be evaluated composed of phenotypes combined by python arithmetic operations.
        return_date: The date to be returned for the phenotype. Can be "first", "last", or a Phenotype object.
        name: The name of the phenotype.

    Attributes:
        table (PhenotypeTable): The resulting phenotype table after filtering (None until execute is called)
    """

    def __init__(
        self,
        expression: ComputationGraph,
        return_date: Union[str, Phenotype] = "first",
        name: str = None,
        **kwargs,
    ):
        super(LogicPhenotype, self).__init__(
            name=name,
            expression=expression,
            return_date=return_date,
            operate_on="boolean",
            populate="boolean",
            reduce=True,
        )



class Reporter:
    """
    A reporter creates an analysis of a cohort. It should receive a cohort, execute and return the report.

    To subclass:
        1. implement execute method, returning a table


    Parameters:
        pretty_display: True by default. If true, dataframe is ready for display in a study report with values rounded to the decimal place and phenotype display names shown. Additionally, numeric values are cast to strings so that empty strings "" are displayed rather than NaNs.
        decimal_places: Number of decimal places to round to. By default set to 1.
    """

    def __init__(self, decimal_places: int = 1, pretty_display: bool = True):
        self.decimal_places = decimal_places
        self.pretty_display = pretty_display

    def execute(self, cohort):
        raise NotImplementedError
